# -*- coding: utf-8 -*-
"""_____________________________________________________________________

:PROJECT: LARAsuite

*generate YAML Schema *

:details: generate_yaml_template - a YAML Schema generator for LARA-django.
          This is a LARA-django-base management command to generate a YAML import template for a
          given app. The generated schema can be used to import data into LARA via the gRPC interface.

          usage:
            lara-django-dev generate_yaml_template <lara_app>

          once the template is generated, it can be used to import data into LARA via the gRPC interface:

:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""

import re
import os
import logging

from django.apps import apps
from django.conf import settings
from django.core.management.base import LabelCommand, CommandError
from django.db import models

# Configurable constants
MAX_LINE_WIDTH = getattr(settings, "MAX_LINE_WIDTH", 120)
INDENT_WIDTH = getattr(settings, "INDENT_WIDTH", 2)

SCHEMA_FILE_HEADER = """#_____________________________________________________________________
#  :PROJECT: LARAsuite

#  {lara_app} YAML data import template

# :details: {lara_app} YAML template for import of model instances into LARA via the gRPC interface.
#            generated with 
#            lara-django-dev generate_yaml_template {lara_app}
# :authors: {lara_author} <{lara_author_email}>
#________________________________________________________________________

"""


def indent(level=1, text="", end="\n"):
    return "\n".join([(level * INDENT_WIDTH * " ") + line for line in text.split("\n")]) + end


# safe_string = lambda s: s.replace('"', '\\"').replace("\n", "\\n")
# safe_string = lambda s: f'"{s}"' if s else ""


class Command(LabelCommand):
    help = """Generate a YAML template for a given LARA app (models)"""
    args = "[app_name, author_name, author_email]"
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument("app_name", help="Name of the LARA app to generate the YAML template for")

        parser.add_argument("--author-name", help="Name of the author", default="lara")
        parser.add_argument("--author-email", help="Email of the author", default="lara@lara.org")
        parser.add_argument("--force", action="store_true", help="overwrite existing files", default=False)

        # parser.add_argument('model_name', nargs='*')

    # @signalcommand
    def handle(self, *args, **options):
        self.app_name = options["app_name"]

        logging.warning("!! only a scaffold is generated, please check/add content to the generated files !!!")

        try:
            app = apps.get_app_config(self.app_name)
        except LookupError:
            self.stderr.write("This command requires an existing app name as argument")
            self.stderr.write("Available apps:")
            app_labels = [app.label for app in apps.get_app_configs()]
            for label in sorted(app_labels):
                self.stderr.write("    %s" % label)
            return

        model_res = []

        YAMLApp(app, model_res, options)


class YAMLApp:
    def __init__(self, app_config, model_res, options):
        self.app_config = app_config
        self.model_res = model_res

        self.app_name = options["app_name"]
        self.force_overwrite = options["force"]
        self.author_name = options["author_name"]
        self.author_email = options["author_email"]

        self.schema_str = ""
        self.model_dict = {}
        self.internal_model_name = set()
        self.external_model_name = set()

        self.model_names = [model.__name__ for model in self.app_config.get_models()]

        self.generate_schema()

    def generate_schema(self):
        self.generate_schema_header()
        self.generate_classes()

        # saving schema to file
        schema_filename = f"{self.app_name}_data_import_template.yaml"

        print(f"writing {schema_filename} to {os.getcwd()}")
        # print(self.schema_str)

        if os.path.isfile(schema_filename):
            if self.force_overwrite:
                with open(schema_filename, "w") as f:
                    f.write(self.schema_str)
            else:
                append = input(
                    f"{schema_filename} already exists,\n press 'a' to append,\n press 'o' to overwrite,\n or 'n' to do nothing: "
                )
                if append.lower() == "a":
                    with open(schema_filename, "a") as f:
                        f.write(self.schema_str)
                elif append.lower() == "o":
                    with open(schema_filename, "w") as f:
                        f.write(self.schema_str)
        else:
            with open(schema_filename, "w") as f:
                f.write(self.schema_str)

    def generate_schema_header(self):
        self.schema_str += SCHEMA_FILE_HEADER.format(
            lara_app=self.app_name,
            lara_author=self.author_name,
            lara_author_email=self.author_email,
        )
        # adding LARA prefixes
        # self.schema_str += "".join([indent(1, f"{prefix}: {uri}") for prefix, uri in settings.LARA_PREFIXES.items()])

        # adding imports
        self.schema_str += """\n"""

    def generate_classes(self):
        class_str = ""
        for model in self.app_config.get_models():
            self.internal_model_name.add(model.__name__)

            class_str += (
                indent(0, "#________  " + model.__name__ + "  ________")
                # + indent(1, "# description:")
                # + indent(1, "#  " +  model.__doc__.replace("\n"," ").strip())
                # + indent(1, "# parameters:")
                + self.parse_fields(model)
                + "\n"
            )

        # first adding external classes
        self.external_model_name = self.external_model_name - self.internal_model_name
        # print(f"external classes: {self.external_classes}"
        # self.schema_str += "".join([indent(1, f"{external_class}:\n") for external_class in self.external_model_name])

        for external_class in self.external_model_name:
            self.schema_str += (
                indent(1, external_class + ":") + indent(2, f"description: {self.model_dict[external_class]}") + "\n"
            )

        # then adding internal classes
        self.schema_str += class_str
        # ext_name = self.external_model_name.pop()
        # print(f"ext classes: {ext_name} -> {self.ext_model_dict[ext_name]}")

    def parse_fields(self, model):
        fields_str = "- "

        default_prefix = "lara"
        indent_level = 0
        cs = 8 # comment_spacing
        for field in model._meta.get_fields():
            #print(f"field: {field} --> {indent_level} \n")
            # print(f"field: {field} --> {type(field)} \n")

            if (
                isinstance(field, models.fields.related.ManyToManyField)
                or isinstance(field, models.fields.reverse_related.ManyToOneRel)
                or isinstance(field, models.fields.reverse_related.ManyToManyRel)
                or isinstance(field, models.fields.reverse_related.OneToOneRel)
                or isinstance(field, models.fields.reverse_related.ManyToManyRel)
            ):
                required = False
            else:
                required = field.blank == False

            match field:
                case models.fields.related.ManyToOneRel():
                    indent_level = 0
                case models.fields.UUIDField():
                    if field.primary_key:
                        indent_level = 0
                        fields_str += indent(indent_level, field.name + f":{' '*cs}#  UUID string (primary key, required)")
                    else:
                        fields_str += indent(indent_level, field.name + f":{' '*cs}#  UUID string ({'required' if required else 'optional'})")
                    # self.external_model_name.add("UUID")

                case models.fields.BooleanField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  boolean ({'required' if required else 'optional'}) ")
                case models.fields.CharField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  string ({'required' if required else 'optional'})")

                case models.fields.TextField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  string ({'required' if required else 'optional'})")
                case models.fields.IntegerField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  integer ({'required' if required else 'optional'})")

                case models.fields.FloatField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  float ({'required' if required else 'optional'})")

                case models.fields.DecimalField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  float ({'required' if required else 'optional'})")

                case models.fields.DateTimeField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  datetime, e.g. 2024-03-31T00:00:00Z ({'required' if required else 'optional'})")

                case models.fields.DateField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  date, e.g. 2024-03-31 ({'required' if required else 'optional'})")
 
                case models.fields.json.JSONField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  json ({'required' if required else 'optional'})")

                case models.fields.files.FileField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  file ({'required' if required else 'optional'})")

                case models.fields.related.ForeignKey():
                    # print(f"{field.name} --- attributes: {field.__dict__}")
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  foreign key ({'required' if required else 'optional'})")
                    self.external_model_name.add(field.related_model.__name__)

                    self.model_dict[field.related_model.__name__] = field.related_model

                case models.OneToOneField():
                    fields_str += indent(indent_level, field.name + f": null{' '*cs}#  one2one key ({'required' if required else 'optional'})")
                    self.external_model_name.add(field.related_model.__name__)

                    self.model_dict[field.related_model.__name__] = field.related_model

                case models.fields.related.ManyToManyField():
                    # print(f"field attributes: {field.__dict__}")
                    fields_str += indent(indent_level, field.name + f": []{' '*cs}#  many2many key ({'required' if required else 'optional'})")
                    self.external_model_name.add(field.related_model.__name__)

                    self.model_dict[field.related_model.__name__] = field.related_model

            indent_level = 1

        return fields_str

    def generate_container(self):
        self.schema_str += indent(1, "Container:") + indent(2, "tree_root: true") + indent(2, "attributes:")

        for model in self.app_config.get_models():
            self.schema_str += (
                indent(3, model.__name__.lower() + "s:")
                + indent(4, "range: " + model.__name__)
                + indent(4, "multivalued: true")
                + indent(4, "inlined_as_list: true")
                + "\n"
            )



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_base admin *

:details: lara_django_base admin module admin backend configuration.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev admin_generator lara_django_base >> admin.py" to update this file
________________________________________________________________________
"""

# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import Address
from .models import City
from .models import Country
from .models import CountrySubdivision
from .models import Currency
from .models import DataType
from .models import ExternalResource
from .models import ExtraData
from .models import GeoLocation
from .models import ItemRole
from .models import ItemStatus
from .models import LARASyncServer
from .models import Location
from .models import MediaType
from .models import Namespace
from .models import PhysicalStateMatter
from .models import PhysicalUnit
from .models import Room
from .models import RoomCategory
from .models import ScientificConstant
from .models import SubdivisionType
from .models import Tag


@admin.register(DataType)
class DataTypeAdmin(admin.ModelAdmin):
    list_display = ("datatype_id", "name", "description")


@admin.register(MediaType)
class MediaTypeAdmin(admin.ModelAdmin):
    list_display = (
        "mediatype_id",
        "name",
        "media_type",
        "def_filename_extension_regex",
        "description",
    )
    search_fields = ("name",)


@admin.register(Namespace)
class NamespaceAdmin(admin.ModelAdmin):
    list_display = (
        "uri",
        "description",
        "namespace_id",
    )
    search_fields = ("uri",)


@admin.register(Tag)
class TagAdmin(admin.ModelAdmin):
    list_display = ("tag", "iri", "tag_id")


@admin.register(ItemRole)
class ItemRoleAdmin(admin.ModelAdmin):
    list_display = ("itemrole_id", "role", )
    search_fields = ("role",)

@admin.register(ItemStatus)
class ItemStatusAdmin(admin.ModelAdmin):
    list_display = ("itemstatus_id", "status", "value", "color", )


@admin.register(ExtraData)
class ExtraDataAdmin(admin.ModelAdmin):
    list_display = (
        "name_full",
        "name",
        "namespace",
        "data_type",
        "iri",
        "data_json",
        "media_type",
        "description",
        "extradata_id",
        "file",
    )
    list_filter = ("name", "namespace", "media_type")


@admin.register(ExternalResource)
class ExternalResourceAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "url",
        "iri",
        "externalresource_id",
    )
    list_filter = ("name", "tags", "url", "iri")
    # raw_id_fields = ()


@admin.register(PhysicalStateMatter)
class PhysicalStateMatterAdmin(admin.ModelAdmin):
    list_display = ("physicalstatematter_id", "name", "description")
    search_fields = ("name",)


@admin.register(PhysicalUnit)
class PhysicalUnitAdmin(admin.ModelAdmin):
    list_display = ("physicalunit_id", "name", "dimension", "description")
    search_fields = ("name",)


@admin.register(ScientificConstant)
class ScientificConstantAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "value",
        "unit",
        "data_json",
        "iri",
        "url_source",
        "url_definition",
        "description",
        "scientificconstant_id",
    )
    list_filter = ("unit",)
    search_fields = ("name",)


@admin.register(Currency)
class CurrencyAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "iri",
        "alpha3code",
        "numeric_code",
        "symbol",
        "exchange_rate_eur",
        "currency_id",
    )
    search_fields = ("name",)


@admin.register(GeoLocation)
class GeoLocationAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "coordinates_dd_lat",
        "coordinates_dd_long",
        "elevation",
        "openstreetmap",
        "googlemap",
        "geolocation_id",
    )
    raw_id_fields = ("data_extra",)
    search_fields = ("name",)


@admin.register(Country)
class CountryAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_local",
        "alpha2code",
        "alpha3code",
        "numeric_code",
        "phone_code",
        "icon",
        "currency",
        "country_id",
    )
    raw_id_fields = ("currency",)
    search_fields = ("name",)


@admin.register(SubdivisionType)
class SubdivisionTypeAdmin(admin.ModelAdmin):
    list_display = ("subdivisiontype_id", "name", "description")
    search_fields = ("name",)


@admin.register(CountrySubdivision)
class CountrySubdivisionAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_local",
        "code",
        "alpha2code",
        "subdivision_type",
        "country",
        "timezone",
        "countrysubdivision_id",
    )
    raw_id_fields = ("subdivision_type", "country")
    search_fields = ("name",)


@admin.register(City)
class CityAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_local",
        "country",
        "subdivision",
        "timezone",
        "city_id",
    )
    list_filter = ("country", "subdivision")
    search_fields = ("name",)


@admin.register(Address)
class AddressAdmin(admin.ModelAdmin):
    list_display = (
        "address_id",
        "building",
        "street",
        "house_number",
        "supplement",
        "city",
        "subdivision",
        "country",
        "postal_code",
        "geolocation",
    )
    list_filter = ("city", "subdivision", "country", "geolocation")
    raw_id_fields = ("data_extra",)


@admin.register(RoomCategory)
class RoomCategoryAdmin(admin.ModelAdmin):
    list_display = ("roomcategory_id", "category", "description")


@admin.register(Room)
class RoomAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "category",
        "number",
        "floor",
        "phone_number",
        "room_map",
        "max_people",
        "area",
        "volume",
        "room_id",
    )
    list_filter = ("category",)
    search_fields = ("name",)


@admin.register(Location)
class LocationAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "barcode",
        "geolocation",
        "address",
        # "building",
        # "floor",
        # "room",
        # "cabinet",
        # "shelf",
        # "drawer",
        # "box",
        # "position",
        "iri",
        "url",
        "description",
        "location_id",
    )
    list_filter = ("geolocation", "address")
    raw_id_fields = ("data_extra",)
    search_fields = ("name",)


@admin.register(LARASyncServer)
class LARASyncServerAdmin(admin.ModelAdmin):
    list_display = (
        "url",
        "port",
        "description",
        "laraserver_id",
    )
    list_filter = ("description", "url")
    search_fields = ("url", "tags")



# generic object factory from https://realpython.com/factory-method-python/ 

class ObjectFactory:
    def __init__(self):
        self._builders = {}

    def register_builder(self, key, builder):
        self._builders[key] = builder

    def create(self, key, **kwargs):
        builder = self._builders.get(key)
        if not builder:
            raise ValueError(key)
        return builder(**kwargs)
    
    def list(self) -> list:
        """ return a list of registered builders"""
        return self._builders.keys()
    
    def has(self, key) -> bool:
        """ check if key is in registered builders, return True or False"""
        return key in self._builders.keys()


#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Thanks to  Pina Merkert (pmk@ct.de, https://github.com/pinae/Nutria-DB )

from django.core.management.base import BaseCommand
from django.db import connection
from django.db.utils import OperationalError
import time


class Command(BaseCommand):
    def handle(self, *args, **options):
        print('Checking the connection to the database.')
        database_connected = False
        while not database_connected:
            try:
                connection.ensure_connection()
                database_connected = True
            except OperationalError:
                print(self.style.ERROR(
                    "Database still unavailable."),
                    end='')
                print(" Waiting 1 second.")
                time.sleep(1)
        print(self.style.SUCCESS('Database available!'))



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_base Model Context Protocol (MCP) handler *

:details: lara_django_base Model Context Protocol (MCP) handler.

:authors: mark doerr <mark.doerr@uni-greifswald.de>

________________________________________________________________________
"""

from mcp_server import ModelQueryToolset

from .models import (
    Address,
    City,
    Country,
    CountrySubdivision,
    Currency,
    DataType,
    ExternalResource,
    ExtraData,
    GeoLocation,
    ItemRole,
    ItemStatus,
    LARAExtension,
    LARASyncServer,
    Location,
    LocationType,
    MediaType,
    Namespace,
    PhysicalStateMatter,
    PhysicalUnit,
    Room,
    RoomCategory,
    ScientificConstant,
    SubdivisionType,
    Tag,
)


class MediaTypeQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `MediaType`."""

    model = MediaType


class DataTypeQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `DataType`."""

    model = DataType


class NamespaceQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Namespace`."""

    model = Namespace


class TagQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Tag`."""

    model = Tag


class ItemRoleQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `ItemRole`."""

    model = ItemRole


class ItemStatusQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `ItemStatus`."""

    model = ItemStatus


class ExtraDataQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `ExtraData`."""

    model = ExtraData


class ExternalResourceQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `ExternalResource`."""

    model = ExternalResource


class PhysicalUnitQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `PhysicalUnit`."""

    model = PhysicalUnit


class PhysicalStateMatterQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `PhysicalStateMatter`."""

    model = PhysicalStateMatter


class ScientificConstantQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `ScientificConstant`."""

    model = ScientificConstant


class CurrencyQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Currency`."""

    model = Currency


class LARAExtensionQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `LARAExtension`."""

    model = LARAExtension


class GeoLocationQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `GeoLocation`."""

    model = GeoLocation


class CountryQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Country`."""

    model = Country


class SubdivisionTypeQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `SubdivisionType`."""

    model = SubdivisionType


class CountrySubdivisionQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `CountrySubdivision`."""

    model = CountrySubdivision


class CityQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `City`."""

    model = City


class LocationTypeQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `LocationType`."""

    model = LocationType


class LocationQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Location`."""

    model = Location


class RoomCategoryQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `RoomCategory`."""

    model = RoomCategory


class RoomQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Room`."""

    model = Room


class AddressQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Address`."""

    model = Address


class LARASyncServerQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `LARASyncServer`."""

    model = LARASyncServer






"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_myproj gRPC serializer*

:details: lara_django_myproj gRPC serializer.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

## generated with django-socio-grpc generateprpcinterface lara_django_base  (LARA-version)

from django_socio_grpc import proto_serializers
from lara_django_base_grpc.v1 import lara_django_base_pb2
from rest_framework.serializers import PrimaryKeyRelatedField, UUIDField

from lara_django_base.models import (
    Address,
    City,
    Country,
    CountrySubdivision,
    Currency,
    DataType,
    ExternalResource,
    ExtraData,
    GeoLocation,
    ItemStatus,
    LARASyncServer,
    Location,
    MediaType,
    Namespace,
    PhysicalStateMatter,
    PhysicalUnit,
    Room,
    RoomCategory,
    ScientificConstant,
    SubdivisionType,
    Tag,
)


class DataTypeProtoSerializer(proto_serializers.ModelProtoSerializer):
    media_type = PrimaryKeyRelatedField(
        queryset=MediaType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )

    class Meta:
        model = DataType
        proto_class = lara_django_base_pb2.DataTypeResponse

        proto_class_list = lara_django_base_pb2.DataTypeListResponse

        fields = "__all__"  # [name', description']


class MediaTypeProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = MediaType
        proto_class = lara_django_base_pb2.MediaTypeResponse

        proto_class_list = lara_django_base_pb2.MediaTypeListResponse

        fields = "__all__"  # [name', media_type', def_filename_extension_regex', description']


class NamespaceProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = Namespace
        proto_class = lara_django_base_pb2.NamespaceResponse

        proto_class_list = lara_django_base_pb2.NamespaceListResponse

        fields = "__all__"


class NamespaceByURIProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = Namespace
        proto_class = lara_django_base_pb2.NamespaceResponse

        proto_class_list = lara_django_base_pb2.NamespaceListResponse

        fields = "__all__"


class TagProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = Tag
        proto_class = lara_django_base_pb2.TagResponse

        proto_class_list = lara_django_base_pb2.TagListResponse

        fields = "__all__"  # [tag']


class ItemStatusProtoSerializer(proto_serializers.ModelProtoSerializer):
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True, many=True
    )
    class Meta:
        model = ItemStatus
        proto_class = lara_django_base_pb2.ItemStatusResponse

        proto_class_list = lara_django_base_pb2.ItemStatusListResponse

        fields = "__all__"  # [status', description']


class ExtraDataProtoSerializer(proto_serializers.ModelProtoSerializer):
    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    data_type = PrimaryKeyRelatedField(
        queryset=DataType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    media_type = PrimaryKeyRelatedField(
        queryset=MediaType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )

    class Meta:
        model = ExtraData
        proto_class = lara_django_base_pb2.ExtraDataResponse

        proto_class_list = lara_django_base_pb2.ExtraDataListResponse

        fields = "__all__"  # [name', namespace', URI', text', XML', JSON', bin', media_type',
        # IRI', URL', description', file', image']


class PhysicalUnitProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = PhysicalUnit
        proto_class = lara_django_base_pb2.PhysicalUnitResponse

        proto_class_list = lara_django_base_pb2.PhysicalUnitListResponse

        fields = "__all__"  # [name', dimension', description']


class ExternalResourceProtoSerializer(proto_serializers.ModelProtoSerializer):
    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    resource_type = PrimaryKeyRelatedField(
        queryset=DataType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True, many=True
    )

    class Meta:
        model = ExternalResource
        proto_class = lara_django_base_pb2.ExternalResourceResponse

        proto_class_list = lara_django_base_pb2.ExternalResourceListResponse

        fields = (
            "__all__"  # [name', media_type', URI', text', XML', JSON', bin', IRI', URL', description', file', image']
        )


class PhysicalStateMatterProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = PhysicalStateMatter
        proto_class = lara_django_base_pb2.PhysicalStateMatterResponse

        proto_class_list = lara_django_base_pb2.PhysicalStateMatterListResponse

        fields = "__all__"  # [name', description']


class ScientificConstantProtoSerializer(proto_serializers.ModelProtoSerializer):
    unit = PrimaryKeyRelatedField(queryset=PhysicalUnit.objects.all(), pk_field=UUIDField(format="hex_verbose"))

    class Meta:
        model = ScientificConstant
        proto_class = lara_django_base_pb2.ScientificConstantResponse

        proto_class_list = lara_django_base_pb2.ScientificConstantListResponse

        fields = "__all__"  # [name', value', unit', JSON', IRI', url_source', url_definition', description']


class CurrencyProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = Currency
        proto_class = lara_django_base_pb2.CurrencyResponse

        proto_class_list = lara_django_base_pb2.CurrencyListResponse

        fields = "__all__"  # [name', IRI', alpha3code', numeric_code', symbol', exchange_rate_eur']


class GeoLocationProtoSerializer(proto_serializers.ModelProtoSerializer):
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
        many=True,
    )

    class Meta:
        model = GeoLocation
        proto_class = lara_django_base_pb2.GeoLocationResponse

        proto_class_list = lara_django_base_pb2.GeoLocationListResponse

        fields = "__all__"  # [name', coordinates_dd_lat', coordinates_dd_long', elevation', openstreetmap', googlemap']


class CountryProtoSerializer(proto_serializers.ModelProtoSerializer):
    currency = PrimaryKeyRelatedField(
        queryset=Currency.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )

    class Meta:
        model = Country
        proto_class = lara_django_base_pb2.CountryResponse

        proto_class_list = lara_django_base_pb2.CountryListResponse

        fields = (
            "__all__"  # [name', name_local', alpha2code', alpha3code', numeric_code', phone_code', icon', currency']
        )


class SubdivisionTypeProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = SubdivisionType
        proto_class = lara_django_base_pb2.SubdivisionTypeResponse

        proto_class_list = lara_django_base_pb2.SubdivisionTypeListResponse

        fields = "__all__"  # [name', description']


class CountrySubdivisionProtoSerializer(proto_serializers.ModelProtoSerializer):
    country = PrimaryKeyRelatedField(queryset=Country.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    subdivision_type = PrimaryKeyRelatedField(
        queryset=SubdivisionType.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )

    class Meta:
        model = CountrySubdivision
        proto_class = lara_django_base_pb2.CountrySubdivisionResponse

        proto_class_list = lara_django_base_pb2.CountrySubdivisionListResponse

        fields = "__all__"  # [name', name_local', code', alpha2code', subdivision_type', country', timezone']


class CityProtoSerializer(proto_serializers.ModelProtoSerializer):
    country = PrimaryKeyRelatedField(queryset=Country.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    subdivision = PrimaryKeyRelatedField(
        queryset=CountrySubdivision.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )

    class Meta:
        model = City
        proto_class = lara_django_base_pb2.CityResponse

        proto_class_list = lara_django_base_pb2.CityListResponse

        fields = "__all__"  # [name', name_local', country', subdivision', timezone']


class AddressProtoSerializer(proto_serializers.ModelProtoSerializer):
    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    city = PrimaryKeyRelatedField(queryset=City.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    subdivision = PrimaryKeyRelatedField(
        queryset=CountrySubdivision.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )
    country = PrimaryKeyRelatedField(
        queryset=Country.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    geolocation = PrimaryKeyRelatedField(
        queryset=GeoLocation.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
        many=True,
    )

    class Meta:
        model = Address
        proto_class = lara_django_base_pb2.AddressResponse

        proto_class_list = lara_django_base_pb2.AddressListResponse

        fields = "__all__"  # [building', street', house_number', supplement', city', subdivision',
        # country', postal_code', geolocation']


class RoomCategoryProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = RoomCategory
        proto_class = lara_django_base_pb2.RoomCategoryResponse

        proto_class_list = lara_django_base_pb2.RoomCategoryListResponse

        fields = "__all__"  # [category', description']


class RoomProtoSerializer(proto_serializers.ModelProtoSerializer):
    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    category = PrimaryKeyRelatedField(
        queryset=RoomCategory.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )

    # data_extra = PrimaryKeyRelatedField(
    #     queryset=ExtraData.objects.all(),
    #     pk_field=UUIDField(format="hex_verbose"),
    #     required=False,
    #     allow_null=True,
    #     many=True,
    # )

    class Meta:
        model = Room
        proto_class = lara_django_base_pb2.RoomResponse

        proto_class_list = lara_django_base_pb2.RoomListResponse

        fields = "__all__"  # [name', category', number', floor', phone_number', address',
        # room_map', max_people', area', volume']


class LocationProtoSerializer(proto_serializers.ModelProtoSerializer):
    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    geolocation = PrimaryKeyRelatedField(
        queryset=GeoLocation.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    address = PrimaryKeyRelatedField(
        queryset=Address.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
        many=True,
    )
    # barcode_json = JSONField(required=False, allow_null=True)

    class Meta:
        model = Location
        proto_class = lara_django_base_pb2.LocationResponse

        proto_class_list = lara_django_base_pb2.LocationListResponse

        fields = "__all__"  # [name', barcode1D', barcode2D', geolocation', address', building',
        # floor', room', cabinet', shelf', drawer', box', position', URI', JSON', IRI', URL', description']


class LARASyncServerProtoSerializer(proto_serializers.ModelProtoSerializer):
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True, many=True
    )

    class Meta:
        model = LARASyncServer
        proto_class = lara_django_base_pb2.LARASyncServerResponse

        proto_class_list = lara_django_base_pb2.LARASyncServerListResponse

        fields = "__all__"  # [name', description', host', port', path', version', protocol']



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_base urls *

:details: lara_django_base urls module.
         - add app specific urls here
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

________________________________________________________________________
"""

from . import views
from django.conf import settings
from django.conf.urls.static import static
from django.urls import path, include
from django.views.generic import TemplateView
from django_socio_grpc.settings import grpc_settings

from . import views

# !! this sets the apps namespace to be used in the template
app_name = "lara_django_base"

# companies and institutions should also be added
# the 'name' attribute is used in templates to address the url independent of the view

urlpatterns = [
    path('', views.select_home_view, name="home"), # views.Home.as_view(), name='index'), # views.select_home_view, name="home"), #
    # All embedded-service UIs share EmbeddedUIView — only the target URL and title differ.
    path('chat/', views.EmbeddedUIView.as_view(embed_url="http://localhost:3033/", title="Chat"), name='chat'),
    path('jupyter-lite/', views.EmbeddedUIView.as_view(embed_url="http://localhost:8089/", title="Jupyter Lite"), name='jupyter_lite_ui'),
    path('sparql/', views.EmbeddedUIView.as_view(embed_url="http://localhost:8890/sparql", title="SPARQL"), name='sparql_ui'),
    path('prefect/', views.EmbeddedUIView.as_view(embed_url="http://localhost:4200/", title="Prefect"), name='prefect_ui'),
    path('zotero/', views.EmbeddedUIView.as_view(embed_url="http://localhost:8099/", title="Zotero"), name='zotero_ui'),
    # influx and grafana
    path('influxdb/', views.EmbeddedUIView.as_view(embed_url="http://localhost:8086/", title="InfluxDB"), name='influxdb_ui'),
    path('grafana/', views.EmbeddedUIView.as_view(embed_url="http://localhost:3003/", title="Grafana"), name='grafana_ui'),
    path('dashboard/', views.Dashboard.as_view(), name='dashboard'),
    path('data-monitor/create/', views.add_data_monitor, name='data_monitor_create'),
    path('apps/', views.AppsSelector.as_view(), name='apps'),
    path('about/', views.AboutView.as_view(), name='base-about'),
    path('contact/', views.ContactView.as_view(), name='base-contact'),
    # htmx partials
    path('partials/drawer-projects/', views.DrawerProjectsPartialView.as_view(), name='drawer_projects'),
    path('partials/apps/', views.AppsSelectorPartialView.as_view(), name='apps_list_partial'),

#     path('addresses/list/', views.AddressesListView.as_view(), name='address-list'),
#     path('addresses/create/', views.AddressCreateView.as_view(),
#          name='address-create'),
#     path('addresses/update/<uuid:pk>', views.AddressUpdateView.as_view(),
#          name='address-update'),
#     path('addresses/delete/<uuid:pk>', views.AddressDeleteView.as_view(),
#          name='address-delete'),
#     path('addresses/<uuid:pk>/', views.AddressDetailView.as_view(),
#          name='address-details'),
    # path('subdir/', include('.urls')),
]

grpc_settings.user_settings["GRPC_HANDLERS"] += [
   "lara_django_base.grpc.handlers.grpc_handlers"]



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_base views *

:details: lara_django_base views module.
         - add app specific urls here
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

________________________________________________________________________
"""

# from configparser import ConfigParser, NoOptionError, NoSectionError # this might be obsolete
from dataclasses import dataclass, field

import environ
from django import __version__ as django_version
from django.apps import apps
from django.conf import settings
from django.contrib.auth.forms import UserCreationForm
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import reverse_lazy
from django.views import generic
from django.views.generic import CreateView, DeleteView, DetailView, TemplateView, UpdateView
from django_tables2 import SingleTableView
from lara_django_people.views import PeopleMenu
from lara_django_projects.models import Project

from . import __version__
from .forms import AddressCreateForm, AddressUpdateForm
from .models import Address
from .tables import AddressTable
from .views_mixins import HtmxResponseMixin


@dataclass
class ItemMenu:
    """Container for simple top-level menu items."""

    menu_items: list[dict] = field(
        default_factory=lambda: [
            {"name": "item1", "path": "lara_django_base:view1-name"},
            {"name": "item2", "path": "lara_django_base:view2-name"},
        ]
    )

@dataclass
class SidebarMenu:
    """Container for sidebar navigation items."""

    menu_items: list[dict] = field(
        default_factory=lambda: [
            {"name": "Home", "path": "/", "icon": "home", "hover_class": "hover:bg-gray-500"},
            {"name": "Dashboard", "path": "/dashboard/", "icon": "dashboard", "hover_class": "hover:bg-gray-500"},
            {"name": "Apps", "path": "/apps/", "icon": "apps", "hover_class": "hover:bg-gray-500"},
            {"name": "Projects", "path": "/projects/", "icon": "projects", "hover_class": "hover:bg-red-500"},
            {"name": "Processes", "path": "/processes/", "icon": "processes", "hover_class": "hover:bg-orange-500"},
            {"name": "Data", "path": "/data/", "icon": "data", "hover_class": "hover:bg-sky-500"},
            {
                "name": "Devices",
                "path": "/material-store/device/list/",
                "icon": "devices",
                "hover_class": "hover:bg-amber-500",
            },
            {
                "name": "Parts",
                "path": "/material-store/part/list/",
                "icon": "parts",
                "hover_class": "hover:bg-amber-500",
            },
            {
                "name": "Labware",
                "path": "/material-store/labware/list/",
                "icon": "labware",
                "hover_class": "hover:bg-amber-500",
            },
            {
                "name": "Substances",
                "path": "/substances/",
                "icon": "substances",
                "hover_class": "hover:bg-teal-500",
            },
            {"name": "Organisms", "path": "/organisms-store/", "icon": "organisms", "hover_class": "hover:bg-lime-500"},
            {"name": "People", "path": "/people/", "icon": "people", "hover_class": "hover:bg-blue-500"},
        ]
    )


class BaseTemplateSettingsMixin:
    """Mixin to provide common template settings for project views."""

    def get_context_data(self, **kwargs):  # noqa: D102
        context = super().get_context_data(**kwargs)
        projects = Project.objects.exclude(parent=None)
        context["projects"] = projects
        context["projects_latest"] = projects.order_by("-datetime_last_modified")  # projects_latest
        context["menu_items"] = []  # ItemMenu().menu_items
        context["app_color"] = "gray"
        context["sidebar_menu_items"] = SidebarMenu().menu_items
        # context["sparql_server_url"] = f"http://{settings.VIRTUOSO_HOST}:{settings.VIRTUOSO_ADMIN_PORT}/sparql/"

        return context


def app_version(app):
    """Return the version of a Django app module when available."""
    try:
        return __import__(app).__version__
    except AttributeError:
        return "version unknown"


env = environ.Env(
    # set casting, default value
    DEBUG=(bool, False)
)


def select_home_view(request):
    """Redirect the current user to their configured home screen."""
    # get the user, who is currently logged in

    current_user = request.user

    # TODO: might be obsolete, as the home screen url is now stored in the user model and not in the session
    # user = {
    #     "current_user": current_user,
    #     "authenticated": current_user.is_authenticated,
    #     "home_screen_url": current_user.home_screen_url,
    # }

    # context = {
    #     "user": user,
    # }

    defaut_home_screen_url = (
        current_user.home_screen_url
        if current_user.home_screen_url is not None and current_user.home_screen_url != ""
        else "lara_django_base:dashboard"
    )

    return redirect(defaut_home_screen_url)


class EmbeddedUIView(BaseTemplateSettingsMixin, TemplateView):
    """
    Generic view that embeds an external service in an iframe.

    Configure per URL via ``as_view(embed_url="http://...", title="My Service")``.
    htmx requests receive only the iframe partial; full-page requests get the
    shell template that triggers the htmx load automatically.
    """

    template_name = "lara_django_base/embedded_ui.html"
    embed_url: str = ""
    title: str = ""

    def get_template_names(self) -> list[str]:
        """
        Return the iframe partial for direct htmx requests, the full shell otherwise.

        hx-boost sets both HX-Request and HX-Boosted; the div's hx-get sets only
        HX-Request.  Boosted navigation must still receive the full shell so that
        the layout (navbar, sidebar, …) is preserved.
        """
        is_htmx = self.request.headers.get("HX-Request")
        is_boosted = self.request.headers.get("HX-Boosted")
        if is_htmx and not is_boosted:
            return ["lara_django_base/partials/iframe.html"]
        return [self.template_name]

    def get_context_data(self, **kwargs) -> dict:
        """Inject embed URL and title into the template context."""
        context = super().get_context_data(**kwargs)
        context["embed_url"] = self.embed_url
        context["title"] = self.title
        return context


class DrawerProjectsPartialView(TemplateView):
    """
    Htmx-only partial that renders the project navigation tree for the drawer.

    Returns HTTP 400 for plain (non-htmx) requests so the endpoint is not
    accidentally used as a full page.
    """

    template_name = "lara_django_base/partials/drawer_projects.html"

    def get(self, request, *args, **kwargs):
        """Enforce htmx-only access and return the project tree partial."""
        if not request.headers.get("HX-Request"):
            from django.http import HttpResponseBadRequest

            return HttpResponseBadRequest("htmx request required")
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs) -> dict:
        """Load the project tree for the drawer partial."""
        context = super().get_context_data(**kwargs)
        context["projects"] = Project.objects.exclude(parent=None)
        return context



def add_data_monitor(request):
    """Handle data monitor form submissions."""
    if request.method == "POST":
        # Handle the form submission
        pass
    return render(request, "404.html")

class Dashboard(HtmxResponseMixin, BaseTemplateSettingsMixin, TemplateView):
    """Dashboard class view."""

    partial_template_name = "lara_django_base/partials/dashboard_content.html"
    page_title = "LARA Dashboard"

    def get_context_data(self, **kwargs):  # noqa: D102
        context = super().get_context_data(**kwargs)
        context["title"] = "Dashboard"
        return context


def _collect_lara_apps() -> list[dict]:
    """Collect and sort all installed LARA app descriptors."""
    lara_apps: list[dict] = []
    for app in apps.get_app_configs():
        try:
            lara_apps.append(
                {
                    "name": app.verbose_name_short,
                    "description": app.lara_app_description,
                    "icon": app.lara_app_icon,
                    "color": app.lara_app_color,
                    "class_apps": app.lara_class_apps,
                    "instance_apps": app.lara_instance_apps,
                }
            )
        except AttributeError:
            pass
    return sorted(lara_apps, key=lambda a: a["name"])


class AppsSelector(HtmxResponseMixin, BaseTemplateSettingsMixin, TemplateView):
    """Apps selector page — cards are lazy-loaded via a nested htmx request."""

    partial_template_name = "lara_django_base/partials/apps_content.html"
    page_title = "LARA Apps"

    def get_context_data(self, **kwargs) -> dict:
        """Return context for the apps page."""
        context = super().get_context_data(**kwargs)
        context["sparql_server_url"] = (
            f"http://{settings.VIRTUOSO_HOST}:{settings.VIRTUOSO_ADMIN_PORT}/sparql/"
        )
        return context


class AppsSelectorPartialView(TemplateView):
    """
    Htmx-only partial that returns the rendered list of app cards.

    Returns HTTP 400 for plain (non-htmx) requests.
    """

    template_name = "lara_django_base/partials/apps_list.html"

    def get(self, request, *args, **kwargs):
        """Enforce htmx-only access and return the apps-card partial."""
        if not request.headers.get("HX-Request"):
            from django.http import HttpResponseBadRequest

            return HttpResponseBadRequest("htmx request required")
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs) -> dict:
        """Load sorted LARA app descriptors for the cards partial."""
        context = super().get_context_data(**kwargs)
        context["apps"] = _collect_lara_apps()
        return context


class Home(TemplateView):
    # class HomeMenu(TemplateView):
    """Home page view."""

    template_name = "lara_django_base/index.html"

    login_url = "accounts/login"

    def get_context_data(self, **kwargs):
        """Return context for the home page."""
        context = super().get_context_data(**kwargs)
        context["menu_items"] = [  # {'name': 'Material', 'path': 'lara_django_material:part-list'},
            #  {'name': 'Devices',
            #      'path': 'lara_django_material:device-list'},
            #  {'name': 'Part',
            #      'path': 'lara_django_material:part-list'},
            #  {'name': 'Labware',
            #   'path': 'lara_django_material:labware-list'},
            #  {'name': 'Substances',
            #   'path': 'lara_django_material:labware-list'}
        ]
        # context['sparql_server_url'] = f'http://{settings.VIRTUOSO_HOST}:{settings.VIRTUOSO_ADMIN_PORT}/sparql/'
        return context


class AboutView(HtmxResponseMixin, BaseTemplateSettingsMixin, TemplateView):
    """About page view."""

    partial_template_name = "lara_django_base/partials/about_content.html"
    page_title = "About LARA"

    def get_context_data(self, **kwargs):
        """Return context for the about page."""
        context = super().get_context_data(**kwargs)

        django_env = env("DJANGO_SETTINGS_MODULE")

        django_envs = {
            "lara_django.settings.development": "DEVELOPMENT",
            "lara_django.settings.production": "PRODUCTION",
            "lara_django.settings.staging": "STAGING",
        }

        try:
            django_environment = django_envs[django_env]
        except KeyError:
            django_environment = "UNKNOWN"

        # context["apps"] =

        context["lara_env"] = {
            "lara_version": __version__,
            "django_version": django_version,
            "lara_description": "The LARA-django is the comprehensive science database for the LARA suite",
            "django_environment": django_environment,
            "django_settings": django_env,
            "lara_authors": ["mark doerr (mark.doerr@uni-greifswald.de)", "benjamin lear"],
            "apps": [{"name": app, "version": app_version(app)} for app in settings.INSTALLED_APPS],
        }
        return context


class ContactView(HtmxResponseMixin, BaseTemplateSettingsMixin, TemplateView):
    """Contact page view."""

    partial_template_name = "lara_django_base/partials/contact_content.html"
    page_title = "Contact"

    def get_context_data(self, **kwargs):
        """Return context for the contact page."""
        context = super().get_context_data(**kwargs)
        return context


class SignUp(generic.CreateView):
    """User signup view."""

    form_class = UserCreationForm
    success_url = reverse_lazy("login")
    template_name = "signup.html"


class AddressDetailView(DetailView):
    """Detail view for a single address."""

    model = Address

    template_name = "lara_django_people/address_detail.html"

    def get_context_data(self, **kwargs):
        """Return context for the address detail page."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Address - Details"
        context["update_link"] = "lara_django_people:address-update"
        context["menu_items"] = PeopleMenu().menu_items
        return context


class AddressesListView(SingleTableView):
    """Table view listing all addresses."""

    model = Address
    table_class = AddressTable

    template_name = "lara_django_people/list.html"

    def get_context_data(self, **kwargs):
        """Return context for the address list page."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Addresses - List"
        context["create_link"] = "lara_django_people:address-create"
        context["menu_items"] = PeopleMenu().menu_items
        return context


class AddressCreateView(CreateView):
    """Create view for a new address."""

    model = Address
    template_name = "lara_django_people/create_form.html"
    form_class = AddressCreateForm
    success_url = "/people/addresses/list"

    def get_context_data(self, **kwargs):
        """Return context for the address creation page."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Address - Creation"
        return context

    def form_valid(self, form):
        """Handle valid form submission for address creation."""
        print("**** form val")
        if "cancel" in self.request.POST:
            return HttpResponseRedirect(self.get_success_url())
        return super().form_valid(form)


class AddressUpdateView(UpdateView):
    """Update view for an existing address."""

    model = Address
    template_name = "lara_django_people/update_form.html"
    form_class = AddressUpdateForm
    success_url = "/people/addresses/list"

    def get_context_data(self, **kwargs):
        """Return context for the address update page."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Address - Update"
        context["delete_link"] = "lara_django_people:address-delete"

        return context


class AddressDeleteView(DeleteView):
    """Delete view for an existing address."""

    model = Address
    template_name = "lara_django_people/delete_form.html"
    success_url = "/people/addresses/list"

    def get_context_data(self, **kwargs):
        """Return context for the address delete page."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Address - Delete"
        context["delete_link"] = "lara_django_people:address-delete"
        return context



"""Htmx-aware view mixins for LARAsuite apps."""

from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.utils.safestring import mark_safe
from django.views.generic.base import TemplateResponseMixin


class HtmxResponseMixin(TemplateResponseMixin):
    """
    Return a partial for htmx requests, auto-wrapped full page otherwise.

    Every view only needs ``partial_template_name``.  For htmx requests the
    partial is rendered directly.  For normal browser requests the partial is
    rendered first, then injected into ``wrapper_template`` via the
    ``inner_content`` context variable — no per-app full-page template needed.

    Attributes:
        partial_template_name: Path to the partial template rendered for every
            request type.
        wrapper_template: Full-page layout that wraps the partial for normal
            (non-htmx) browser requests.
        page_title: Value passed to the wrapper's ``<title>`` block.

    Example::

        class ProjectListView(HtmxResponseMixin, ListView):
            partial_template_name = "myapp/partials/list.html"
            page_title = "Projects"

    """

    partial_template_name: str | None = None
    wrapper_template: str = "lara_django_base/htmx_wrapper.html"
    page_title: str = ""

    def get_template_names(self) -> list[str]:
        """Always render the partial — wrapping happens in *render_to_response*."""
        if self.partial_template_name:
            return [self.partial_template_name]
        return super().get_template_names()

    def render_to_response(self, context: dict, **response_kwargs) -> TemplateResponse:
        """Wrap partial in full layout for non-htmx requests."""
        if getattr(self.request, "htmx", False):
            # htmx request → return bare partial
            return super().render_to_response(context, **response_kwargs)

        # Normal request → render partial, inject into wrapper
        partial_html = render_to_string(
            self.get_template_names()[0],
            context=context,
            request=self.request,
        )
        wrapper_context = {
            **context,
            "inner_content": mark_safe(partial_html),  # noqa: S308
            "page_title": self.page_title,
        }
        return TemplateResponse(
            self.request,
            self.wrapper_template,
            wrapper_context,
            **response_kwargs,
        )



# s. https://stackoverflow.com/questions/42826287/model-description-in-django-admin

from django import template
from django.utils.html import mark_safe

register = template.Library()


@register.simple_tag()
def model_desc(obj):
    if obj.__doc__:
        return mark_safe('<p>{}</p>'.format(obj.__doc__))
    return ''



import io
import logging

import grpc
from asgiref.sync import sync_to_async
from celery.result import AsyncResult
from django_socio_grpc import mixins
from django_socio_grpc.decorators import grpc_action

logger = logging.getLogger(__name__)

from django_socio_grpc.grpc_actions.actions import GRPCActionMixin
from django_socio_grpc.grpc_actions.placeholders import (
    FnPlaceholder,
    LookupField,
    SelfSerializer,
    StrTemplatePlaceholder,
)
from django_socio_grpc.grpc_actions.utils import get_serializer_base_name
from django_socio_grpc.utils.constants import DEFAULT_LIST_FIELD_NAME, PARTIAL_UPDATE_FIELD_NAME, REQUEST_SUFFIX


from ..tasks import extract_metadata_from_parquet_file


class AsyncUploadModelMixin(GRPCActionMixin):
    @grpc_action(
        request=[
            {"name": "filename", "type": "string"},
            {"name": "update_item_id", "type": "string"},
            {"name": "data", "type": "bytes"},
        ],
        request_name="FileUploadChunk",
        response=[{"name": "success", "type": "bool"}],
        response_name="FileUploadInfo",
        request_stream=True,
    )
    async def UploadFile(self, request, context):
        # logger.info(f"File upload started:")

        filename = ""
        try:
            with io.BytesIO() as f:
                async for file_chunk in request:
                    filename = file_chunk.filename
                    # logger.info("Reading file chunk:", len(file_chunk.data))
                    f.write(file_chunk.data)

                queryset = self.get_queryset()

                item = await queryset.filter(pk=file_chunk.update_item_id).aget()
                # logger.info("--- data found:", data.name_display)
                res = await sync_to_async(item.file.save)(filename, f, save=True)

                f.seek(0)
                file_content = f.getvalue()
                if filename.endswith(".parquet"):
                    task = extract_metadata_from_parquet_file.delay(parquet_file_str=file_content)
                    #logger.info(f"+++ mixin - metadata task: {metadata}")
                    # task_id = task.id
                    # item.metadata = AsyncResult(task_id)  #metadata.collect()
                    # await item.asave()

            logger.info(f"File upload successful: {len(file_content)} bytes")
            return self.pb2.FileUploadInfo(success=True)

        except Exception as e:
            logger.exception(f"Document ({filename}) upload has failed {e}")
            return self.pb2.FileUploadInfo(success=False)

    # image upload
    @grpc_action(
        request=[
            {"name": "filename", "type": "string"},
            {"name": "update_item_id", "type": "string"},
            {"name": "data", "type": "bytes"},
        ],
        request_name="ImageUploadChunk",
        response=[{"name": "success", "type": "bool"}],
        response_name="ImageUploadInfo",
        request_stream=True,
    )
    async def UploadImage(self, request, context):
        logger.info("File upload started:")

        try:
            with io.BytesIO() as f:
                async for file_chunk in request:
                    filename = file_chunk.filename
                    #logger.info("Reading file chunk:", len(file_chunk.data))
                    f.write(file_chunk.data)

                queryset = self.get_queryset()

                item = await queryset.filter(pk=file_chunk.update_item_id).aget()
                #logger.info("--- upload item found:", file_chunk.update_item_id, item.name_display)
                res = await sync_to_async(item.image.save)(filename, f, save=True)

                f.seek(0)
                file_content = f.getvalue()

            logger.info(f"Image File upload successful: {len(file_content)} bytes")
            return self.pb2.ImageUploadInfo(success=True)

        except Exception as e:
            logger.exception(f"Image File ({filename}) upload has failed {e}")
            return self.pb2.ImageUploadInfo(success=True)


class AsyncRetrieveByNameModelMixin(GRPCActionMixin):
    @grpc_action(
        #request=LookupField,
        request_name=StrTemplatePlaceholder(
            f"{{}}RetrieveByNameNS{REQUEST_SUFFIX}", get_serializer_base_name
        ),
        request=[
            {"name": "name", "type": "string"},
            {"name": "namespace_uri", "type": "string"},
        ],
        response=SelfSerializer,
    )
    async def RetrieveByNameNamespace(self, request, context):
        """
        Retrieve a model instance.

        The request have to include a field corresponding to
        ``lookup_request_field``.  If an object can be retrieved this returns
        a proto message of ``serializer.Meta.proto_class``.
        """
        queryset = self.get_queryset()
        queryset = await self.afilter_queryset(queryset)

        if request.namespace_uri is None or request.namespace_uri == "":
            queryset = await sync_to_async(queryset.filter)(name=request.name)
        else:
            queryset = await sync_to_async(queryset.filter)(namespace__uri=request.namespace_uri, name=request.name)

        instance = await queryset.afirst()

        serializer = await self.aget_serializer(instance)
        return await serializer.amessage

    @staticmethod
    def get_default_method(model_name):
        return {
            "Retrieve": {
                "request": {"is_stream": False, "message": f"{model_name}RetrieveRequest"},
                "response": {"is_stream": False, "message": model_name},
            },
        }

    @staticmethod
    def get_default_message(model_name, fields="__pk__"):
        return {
            f"{model_name}RetrieveRequest": fields,
        }

class AsyncDownloadModelMixin(mixins.GRPCActionMixin):
    @grpc_action(
        request=[{"name": "item_id", "type": "string"}, {"name": "chunk_size", "type": "uint64"}],
        request_name="FileDownloadInfo",
        response=[
            {"name": "filename", "type": "string"},
            {"name": "data", "type": "bytes"},
            {"name": "success", "type": "bool"},
        ],
        response_name="FileDownloadChunk",
        response_stream=True,
    )
    async def DownloadFile(self, request, context):
        CHUNK_SIZE = 1024 * 1024  # 1 MB

        logger.info("File download started:")

        filename = ""

        queryset = self.get_queryset()
        instance = await queryset.filter(pk=request.item_id).aget()
        file_field = instance.file
        filename = file_field.name

        try:
            with file_field.open("rb") as f:
                while True:
                    chunk = f.read(CHUNK_SIZE)
                    if not chunk:
                        break
                    yield self.pb2.FileDownloadChunk(filename=filename, data=chunk, success=True)

            logger.info(f"File download successful: {filename}")
            #return self.pb2.FileDownloadInfo(success=True)
        except self.model.DoesNotExist:
            context.abort(grpc.StatusCode.NOT_FOUND, "File not found.")
        except Exception as e:
            logger.exception(f"Document ({filename}) upload has failed {e}")
            context.abort(grpc.StatusCode.INTERNAL, str(e))
            yield self.pb2.FileDownloadInfo(success=False)


    @grpc_action(
        request=[{"name": "item_id", "type": "string"},
                 {"name": "chunk_size", "type": "uint64"}],
        request_name="ImageDownloadInfo",
        response=[
            {"name": "filename_image", "type": "string"},
            {"name": "data", "type": "bytes"},
            {"name": "success", "type": "bool"},
        ],
        response_name="ImageDownloadChunk",
        response_stream=True,
    )
    async def DownloadImage(self, request, context):
        CHUNK_SIZE = 1024 * 1024  # 1 MB

        logger.info("File download started:")

        filename = ""

        queryset = self.get_queryset()

        instance = await queryset.filter(pk=request.item_id).aget()
        file_field = instance.image
        filename = file_field.name

        try:
            with file_field.open("rb") as f:
                while True:
                    chunk = f.read(CHUNK_SIZE)
                    if not chunk:
                        break
                    yield self.pb2.ImageDownloadChunk(filename_image=filename, data=chunk, success=True)
            
            logger.info(f"Image download successful: {filename}")
            # return self.pb2.ImageDownloadInfo(success=True)
        except self.model.DoesNotExist:
            context.abort(grpc.StatusCode.NOT_FOUND, "Image not found.")
        except Exception as e:
            logger.exception(f"Image ({filename}) download has failed {e}")
            context.abort(grpc.StatusCode.INTERNAL, str(e))
            # return self.pb2.ImageDownloadInfo(success=False)
            



from .models_base import (
    BlockchainHashesAbstr,
    Currency,
    DataType,
    ExternalResource,
    ExtraData,
    ExtraDataAbstr,
    ItemRole,
    ItemStatus,
    LARAExtension,
    MediaType,
    Namespace,
    PhysicalStateMatter,
    PhysicalUnit,
    ScientificConstant,
    Tag,
    URNField,
)
from .models_location import (
    Address,
    City,
    Country,
    CountrySubdivision,
    GeoLocation,
    Location,
    LocationType,
    Room,
    RoomCategory,
    SubdivisionType,
)
from .models_sync_calendar import (
    CalendarAbstr,
    LARASyncServer,
    SynchronisationAbstr,
    SyncStatusAbstr,
)

__all__ = [
    "Address",
    "BlockchainHashesAbstr",
    "CalendarAbstr",
    "City",
    "Country",
    "CountrySubdivision",
    "Currency",
    "DataType",
    "ExternalResource",
    "ExtraData",
    "ExtraDataAbstr",
    "GeoLocation",
    "ItemRole",
    "ItemStatus",
    "LARAExtension",
    "LARASyncServer",
    "Location",
    "LocationType",
    "MediaType",
    "Namespace",
    "PhysicalStateMatter",
    "PhysicalUnit",
    "Room",
    "RoomCategory",
    "ScientificConstant",
    "SubdivisionType",
    "SyncStatusAbstr",
    "SynchronisationAbstr",
    "Tag",
    "URNField",
]



"""
LARA Django project initialization.

_____________________________________________________________________.

:PROJECT: lara-django

*lara-django project intitialisation*

:details: initialise lara-django project (run this only once !)
          this file is highly inspired by
          https://gitlab.com/mayan-edms/mayan-edms/-/blob/master/mayan/apps/common/management/commands/initialsetup.py


:authors: mark doerr (mark@uni-greifswald.de)

:date: (creation)          20200409

.. note:: -
.. todo:: -
________________________________________________________________________
"""


import logging

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS, connections

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Django management command to initialize the LARA project.

    See https://docs.djangoproject.com/en/1.9/howto/custom-management-commands/ for more details.
    Uses the argparse mechanism introduced in Django > 1.8.
    """

    # logging.basicConfig(format='%(levelname)s| %(module)s.%(funcName)s:%(message)s',
    #                     level=logging.DEBUG)  # level=logging.ERROR

    help = 'Initialise the lara-django project with default settings - run only once.'

    def add_arguments(self, parser):
        """Add command-line arguments for this command."""
        # defining named commandline arguments
        parser.add_argument('-c', '--collect-static',
                            action='store_true',
                            help='collect static files into static dir')
        parser.add_argument('-n', '--no-fixtures',
                            action='store_true',
                            help='do not load fixtures')
        parser.add_argument('-d', '--load-demo-data',
                            action='store_true',
                            help='load demo data')
        parser.add_argument(
            '--force', action='store_true',
            help='Force execution of the initialization process.',
        )
        parser.add_argument('--make-all-migrations',
                            action='store_true',
                            help='make all migrations, but do not reset and initialise database')
        parser.add_argument('-f', '--load-fixtures',
                            action='store_true',
                            help='load fixtures')
        # list all apps with versions
        parser.add_argument( '-l',
            '--list-apps', action='store_true',
            help='List all installed apps with their versions.',
        )

        # should be removed
        # parser.add_argument(
        #    '--no-dependencies', action='store_true', dest='no_dependencies',
        #    help='Don\'t download dependencies.',
        # )

    def handle(self, *args, **options):
        """Dispatch based on command-line arguments/options."""
        # check if database is initialised (i.e. auth_user table exists)
        connection = connections[DEFAULT_DB_ALIAS]
        table_name = "auth_group"
        db_initialised = table_name in connection.introspection.table_names()

        logger.info(f"LARA db initialised: {db_initialised}")

        if options['make_all_migrations']:
            self.make_app_migrations(options)
        elif options['load_fixtures']:
            self.load_fixtures(options)
        elif options['list_apps']:
            self.list_apps()
        else:
            logger.info("initialising lara-django project database....")

            if db_initialised and not options.get('force', False):
                logger.info("database already initialised, waiting for db....")
                call_command('wait_for_db')
                return
            else:
                self.initialise_lara_django(force=options.get('force', False))
            # pre_initial_setup.send(sender=self)

            if not options.get('no_dependencies', False):
                # call_command(command_name='install_dependencies')
                # call_command(
                #    command_name='prepare_static', interactive=False
                # )
                pass

        if options['collect_static']:
            call_command('collectstatic', '--noinput')  # interactive=False


        if not options['no_fixtures'] and not db_initialised:
            self.load_fixtures(options)

            # now create the admin / superuser
            call_command('init_admin')

        if options['load_demo_data']:
            self.load_demo_data()

    def initialise_lara_django(self, force=False):
        """
        Initialise lara-django.

        Parameters
        ----------
        force : bool, optional
            Force overwriting of all files.

        """
        # s. mayan
        #system_path = os.path.join(settings.MEDIA_ROOT, SYSTEM_DIR)
        #settings_path = os.path.join(settings.MEDIA_ROOT, 'hellodjango_settings')
        #secret_key_file_path = os.path.join(system_path, SECRET_KEY_FILENAME)

        call_command('wait_for_db')

        call_command('makemigrations')

        if self.num_migrations() > 0:
            logger.debug(f"unresolved {self.num_migrations()} migrations found")
            self.make_app_migrations()

        call_command('migrate')

    def num_migrations(self):
        """Return the number of pending migrations."""
        from django.db.migrations.executor import MigrationExecutor
        try:
            executor = MigrationExecutor(connections[DEFAULT_DB_ALIAS])
        except ImproperlyConfigured:
            # No databases are configured (or the dummy one)
            return

        plan = executor.migration_plan(executor.loader.graph.leaf_nodes())
        if plan:
            return len(plan)
        else:
            return 0

    def make_app_migrations(self, options=None):
        """
        Make individual app migrations.

        See https://docs.djangoproject.com/en/2.0/ref/django-admin/.
        """
        #db_filename = settings.DATABASES['default']['NAME']
        for app in settings.LARA_APPS:
            logger.debug(f"migrating : {app}")
            call_command('makemigrations', app)
        call_command('migrate')

    def load_fixtures(self, options=None):
        """
        Load fixtures into the LARA database.

        Parameters
        ----------
        options : dict | None
            Command options.

        """
        logger.info("initialising database with fixtures....")
        for fixture in settings.FIXTURES:
            call_command('loaddata', fixture)

    def load_demo_data(self, options=None):
        """
        Load demo data into the LARA database.

        Parameters
        ----------
        options : dict | None
            Command options.

        """
        logger.info("loading demo data....")
        if settings.DEMO_DATA is not None:
            for fixture in settings.DEMO_DATA:
                call_command('loaddata', fixture)

    def list_apps(self):
        """List all installed apps with their versions."""
        logger.info("listing all installed apps with their versions....")
        for app in settings.INSTALLED_APPS:
            try:
                module = __import__(app, fromlist=['__version__'])
                version = getattr(module, '__version__', 'No version found')
            except ImportError:
                version = 'No version found'
            logger.info(f"{app}: {version}")



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_base celery tasks*

:details: lara_django_base celery tasks.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from celery import shared_task
from celery.utils.log import get_task_logger
from django.apps import apps
from rdflib import Graph
from rdflib import URIRef
# from scidats import SciDatS

logger = get_task_logger(__name__)

def insert_query(query):
    """Insert metadata into triple store using SPARQL."""
    sparql = apps.get_app_config("lara_django_base").sparql

    sparql.setMethod("POST")

    try:
        sparql.setQuery(query)
        qres = sparql.query().convert()
        logger.debug("+++ insert result: %s", qres)

    except Exception:
        logger.exception("Exception occurred while inserting query: %s", query)


@shared_task
def semantic_data_to_triplestore(instance_name: str | None = None,
                                 data_model: str | None = None,
                                 attributes: dict | None = None,
                                 graph_urn: str | None = None):
    """
    Semantic data to triplestore celery task.

    Metadata is converted to semantic data (= owlready2 data) object,
    then triples are extracted from RDF graph
    and written to triple store using SPARQL.
    """
    # insert_semantics(sender, instance, **kwargs)
    # print(f"semantic_data_to_triplestore: {current_task.request.id}")
    # print(f"semantic_data_to_triplestore: {attributes}")

    lara_onto = apps.get_app_config("lara_django_base").lara_onto
    lara_onto_graph = apps.get_app_config("lara_django_base").lara_onto_graph

    data_model_class = getattr(lara_onto, data_model, None)
    if data_model_class is None:
        logger.error("Data model %s not found in lara_onto.", data_model)
        return
    semantic_data_class_instance = data_model_class(instance_name, **attributes)

    triples_str = ""
    for triple in lara_onto_graph.triples((URIRef(semantic_data_class_instance.iri), None, None)):
        triples_str += (
            triple[0].n3() + " " + triple[1].n3() + " " + triple[2].n3() + " .\n"
        )

    query = f"""INSERT INTO <{graph_urn}> {{ {triples_str} }}"""

    # logger.debug(f"+++ insert query:\n {query}")

    insert_query(query)


@shared_task
def extract_metadata_from_parquet_file(parquet_file_str : bytes | None = None) :
   """
      Extract metadata from parquet file.

      1. read parquet file
      2. extract metadata
      3. write metadata to triple store
      4. return metadata
   """
   graph_urn = "urn:sparql:lara:data"

   logger.debug("+++ BASE - extract_metadata_from_parquet_file:")
   return

   data_scidats = None #SciDatS()
   data_scidats.read_parquet(input_stream=parquet_file_str, base64encoding=False)

   # logger.debug("****** scidatas metadata:", data_scidats.metadata_core)

   # use rdflib to convert JSON-LD to RDF
   g = Graph()
   g.parse(data=data_scidats.metadata_jsonld, format="json-ld")

   # insert metadata into triple store using SPARQL
   logger.debug("****** insert metadata into triple store using SPARQL")

   #triples_str = g.serialize(format="turtle")
   #query = f"""INSERT INTO <{graph_urn}> {{ {triples_str} }}"""

   query = f"INSERT INTO <{graph_urn}> {{\n"
   for s, p, o in g:
      query += f"""    <{s}> <{p}> {f'<{o}>' if isinstance(o, URIRef) else f'"{o}"^^<{o.datatype}>' if hasattr(o, 'datatype') else f'"{o}"'} .\n"""  # noqa: E501
   query += "}"

   logger.debug("+++ insert query: \n%s\n", query)

   insert_query(query)

   return data_scidats.metadata_jsonld



# -*- coding: utf-8 -*-
"""_____________________________________________________________________

:PROJECT: LARAsuite

*generate High Level Interface *

:details: generate_high_level_interface - a High Level Interface generator for LARA-django.
          This is a LARA-django-base management command to generate a High Level Interface for a
          given app. The generated interface can be used to generate various validators.
          s. https://linkml.io/linkml/index.html

          usage:
            lara-django-dev generate_high_level_interface <lara_app>

          once the interface is generated, it can be used to generate a pydantic model (from the LinkML package) with:
            gen-pydantic <lara_app>_interface.yaml   

:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - async 
          - delete: return no list.
________________________________________________________________________
"""

import re
import os
import logging

from django.apps import apps
from django.conf import settings
from django.core.management.base import LabelCommand, CommandError
from django.db import models

# Configurable constants
MAX_LINE_WIDTH = getattr(settings, "MAX_LINE_WIDTH", 120)
INDENT_WIDTH = getattr(settings, "INDENT_WIDTH", 2)

INTERFACE_FILE_HEADER = """\"\"\"_____________________________________________________________________

:PROJECT: LARAsuite

*{lara_app} high_level_interface *

:details: {lara_app} high_level_interface.
         - generated with 'lara-django-dev generate_high_level_interface {lara_app}
:authors: {lara_author} <{lara_author_email}>

.. note:: -
.. todo:: - 
________________________________________________________________________
\"\"\""""

def indent(level=1, text="", end="\n"):
    return "\n".join([(level * INDENT_WIDTH * " ") + line for line in text.split("\n")]) + end

def lower_strip(s):
    return s.lower().replace(" ", "").strip()

class Command(LabelCommand):
    help = """Generate a High Level Interface for a given LARA app (models)"""
    args = "[app_name, author_name, author_email]"
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument("app_name", help="Name of the LARA app to generate the LinML Schema for")

        parser.add_argument("--author-name", help="Name of the author", default="lara")
        parser.add_argument("--author-email", help="Email of the author", default="lara@lara.org")
        parser.add_argument("--force", action="store_true", help="overwrite existing files", default=False)
        # parser.add_argument('model_name', nargs='*')

    # @signalcommand
    def handle(self, *args, **options):
        self.app_name = options["app_name"]

        logging.warning("!! only a scaffold is generated, please check/add content to the generated files !!!")

        try:
            app = apps.get_app_config(self.app_name)
        except LookupError:
            self.stderr.write("This command requires an existing app name as argument")
            self.stderr.write("Available apps:")
            app_labels = [app.label for app in apps.get_app_configs()]
            for label in sorted(app_labels):
                self.stderr.write("    %s" % label)
            return

        model_res = []
        # for arg in options['model_name']:
        #     model_res.append(re.compile(arg, re.IGNORECASE))

        HigLevelInterfaceApp(app, model_res, options)

class HigLevelInterfaceApp:
    def __init__(self, app_config, model_res, options):
        self.app_config = app_config
        self.model_res = model_res

        self.app_name = options["app_name"]
        self.base_app_name = self.app_name[len('lara_django_'):] #self.app_name.split("_")[2]
        self.force_overwrite = options["force"]
        self.author_name = options["author_name"]
        self.author_email = options["author_email"]

        self.interface_str = ""
        self.model_names = [model.__name__ for model in self.app_config.get_models()]

        self.generate_interface()

    def generate_interface(self):
        self.generate_interface_header()
        self.generate_interface_imports()
        self.generate_classes()
       
        # saving interface to file
        interface_filename = f"{self.app_name}_interface.py"

        print(f"writing {interface_filename} to {os.getcwd()}")
        #print(self.interface_str)

        if os.path.isfile(interface_filename):
            if self.force_overwrite:
                with open(interface_filename, "w") as f:
                    f.write(self.interface_str)
            else:
                append = input(
                    f"{interface_filename} already exists,\n press 'a' to append,\n press 'o' to overwrite,\n or 'n' to do nothing: "
                )
                if append.lower() == "a":
                    with open(interface_filename, "a") as f:
                        f.write(self.interface_str)
                elif append.lower() == "o":
                    with open(interface_filename, "w") as f:
                        f.write(self.interface_str)
        else:
            with open(interface_filename, "w") as f:
                f.write(self.interface_str)

    def generate_interface_header(self):
        self.interface_str += INTERFACE_FILE_HEADER.format(
            lara_app=self.app_name,
            lara_author=self.author_name,
            lara_author_email=self.author_email,
        )

    def generate_interface_imports(self):
                
        self.interface_str += f"""

import json
import asyncio
import logging
import grpc
from typing import Union

from google.protobuf.json_format import MessageToDict

import lara_django_base_grpc.v1.lara_django_base_pb2 as lara_django_base_pb2
import lara_django_base_grpc.v1.lara_django_base_pb2_grpc as lara_django_base_pb2_grpc

import {self.app_name}_grpc.v1.{self.app_name}_pb2 as {self.app_name}_pb2
import {self.app_name}_grpc.v1.{self.app_name}_pb2_grpc as {self.app_name}_pb2_grpc

import  lara_django_base_grpc.lara_django_base_data_model as base_dm
import  lara_django_base_grpc.lara_django_base_interface as base_if
from  lara_django_base_grpc.abstract_interface import LARAInterface, GRPCChannelSingleton, import_model_instance_from_file, export_model_instance_to_file, upload_file, download_file, update_file, upload_image, update_image, download_image, id_cache

import  {self.app_name}_grpc.{self.app_name}_data_model as {self.base_app_name}_dm 

logger = logging.getLogger(__name__)

\n\n"""


    def generate_classes(self):
        for model in self.app_config.get_models():

            has_namespace = False
            model_field_names = [field.name for field in model._meta.get_fields()]
            
            if "namespace" in model_field_names:
                has_namespace = True

            self.interface_str += f"""#_________________  {model.__name__}  ______________________\n\n
class {model.__name__}Interface:
    def __init__(self, channel: grpc.Channel = None, default_namespace_uri : str = None) -> None:
        self._default_namespace_uri = None
        self.default_namespace_id = None

        if default_namespace_uri is not None:
            self._default_namespace_uri = default_namespace_uri

        if channel is None:
            cs = GRPCChannelSingleton()
            channel = cs.channel
        
        self.namespace_if = base_if.NamespaceInterface(channel)

        # self.{model.__name__.lower()}_class_client = (
        #     {self.app_name}_pb2_grpc.{model.__name__}classByIRIControllerStub(channel)
        # )
        
        self.{model.__name__.lower()}_client = {self.app_name}_pb2_grpc.{model.__name__}ControllerStub(channel)

        # this is required by the upload_file decorator, please activate if needed for file upload
        # self.stub = self.{model.__name__.lower()}_client
        # self.file_chunk = {self.app_name}_pb2.FileChunk
        # self.update_item_id = "{model.__name__.lower()}_id"

        # self.{model.__name__.lower()}_full_client = {self.app_name}_pb2_grpc.{model.__name__}FullControllerStub(
        #     channel
        # )

    @property
    def default_namespace_uri(self):
        return self._default_namespace_uri

    @default_namespace_uri.setter
    def default_namespace_uri(self, namespace_uri: str = None):
        if namespace_uri is not None:
            self._default_namespace_uri = namespace_uri

    @import_model_instance_from_file(relation_resolver=lambda x: x)
    #@upload_file  # activate if needed for file upload
    #@upload_image  # activate if needed for image upload
    async def create(self,  {lower_strip( model._meta.verbose_name_plural)}:  list[{self.base_app_name}_dm.{model.__name__}] = None,  ids_only: bool = False, namespace_uri: str = None) -> Union[list[str], list[{self.base_app_name}_dm.{model.__name__}]]:
        # TODO: only retrieve default_namespace_id if not already retrieved and namespace_uri is None
        self.namespace_id = None
        # {model.__name__.lower()}class_id = None
        if namespace_uri is None and self._default_namespace_uri is not None:
            namespace_uri = self._default_namespace_uri
        if namespace_uri is None:
            raise ValueError("No namespace URI is given.")
        try:
            self.namespace_id = await self.namespace_if.retrieve_id(namespace_uri=namespace_uri)
        except Exception as e:
            logger.error(f"Error retrieving namespace_id: {{e}}")
        
        # some sample code for retrieving a class IRI and adding namespace and class to the model
        # if {model.__name__.lower()}_class_iri is None:
        #     raise ValueError("No sample class IRI is given.")
        # try:
        #     res = await self.{model.__name__.lower()}_class_client.Retrieve(
        #         {self.app_name}_pb2.{model.__name__}classRetrieveRequest(iri={model.__name__.lower()}_class_iri)
        #     )
        #     {model.__name__.lower()}class_id = res.{model.__name__.lower()}class_id 
        # except Exception as e:
        #     logger.error(f"Error retrieving {model.__name__.lower()}class_id: {{e}}")
        # for {model.__name__.lower()} in {model.__name__.lower()}s:
        #     {model.__name__.lower()}.namespace = self.namespace_id
        #     {model.__name__.lower()}.{model.__name__.lower()}_class = {model.__name__.lower()}class_id

        try:
            create_coroutines = ( self.{model.__name__.lower()}_client.Create({self.app_name}_pb2.{model.__name__}Request(**{model.__name__.lower()}.model_dump())) for {model.__name__.lower()} in {lower_strip( model._meta.verbose_name_plural)} )
            res = await asyncio.gather(*create_coroutines)
            if ids_only:
                return [item.{model.__name__.lower()}_id for item in res]
            else:
                return res

        except Exception as e:
            logger.error(f"Error creating {model.__name__.lower()}: {{e}}")

    @export_model_instance_to_file
    async def retrieve(
        self, 
        {model.__name__.lower()}_id: str = None,
        {model.__name__.lower()}_name: str = None,
        #{model.__name__.lower()}_name_full: str = None,
        filter_dict: dict = None,
        raw : bool = False
        ) -> list[{self.base_app_name}_dm.{model.__name__}]:
        \"\"\" retrieve a list of {model.__name__.lower()} instances by {model.__name__.lower()}_id, {model.__name__.lower()}_name or filter_dict, 
               returns a list of {model.__name__.lower()} data model objects,
               if raw is True, the raw protobuf objects are returned 
        \"\"\"  
        filter_list = []
        results_list = []
        if  {model.__name__.lower()}_id is not None:
            try:
                res =  await self.{model.__name__.lower()}_client.Retrieve(
                    {self.app_name}_pb2.{model.__name__}RetrieveRequest({model.__name__.lower()}_id={model.__name__.lower()}_id) )
                if raw:
                    results_list.append(res)
                else:
                    results_list.append( {self.base_app_name}_dm.{model.__name__}(**MessageToDict(res, preserving_proto_field_name=True)) )
            except Exception as e:
                logger.error(f"Error retrieving {model.__name__.lower()}_id: {{e}}")

        if {model.__name__.lower()}_name is not None:
            filter_list.append({{"name": {model.__name__.lower()}_name}})
        # if {model.__name__.lower()}_name_full is not None:
        #     filter_list.append({{"name_full": {model.__name__.lower()}_name_full}})
        if filter_dict is not None:
            filter_list.append(filter_dict)

        for filter_dict in filter_list:
            metadata = (("filters", (json.dumps(filter_dict))),)
            try:
                res = await self.{model.__name__.lower()}_client.List(
                        {self.app_name}_pb2.{model.__name__}ListRequest(), metadata=metadata
                      )
                if raw:
                      results_list += res.results
                else:
                    # a queue could be used here to speed up the retrieval
                    results = res.results
                    results_list += [ {self.base_app_name}_dm.{model.__name__}(**MessageToDict(res, preserving_proto_field_name=True)) for res in results ]
            except Exception as e:
                logger.error(f"Error retrieving {model.__name__.lower()}_name: {{e}}")

        return results_list

    #@id_cache
    async def retrieve_id(self, 
                          {model.__name__.lower()}_name: str = None,
                          {model.__name__.lower()}_name_full: str = None,
                          iri: str = None,) -> str:
        \"\"\" retrieve the {model.__name__.lower()}_id for a given {model.__name__.lower()}_name, 
               {model.__name__.lower()}_name_full or iri 
        \"\"\"
        if {model.__name__.lower()}_name is not None:
            filter_dict = {{"name": {model.__name__.lower()}_name}}
        elif {model.__name__.lower()}_name_full is not None:
            filter_dict = {{"name_full": {model.__name__.lower()}_name_full}}
        elif iri is not None:
            filter_dict = {{"iri": iri}}
        metadata = (("filters", (json.dumps(filter_dict))),)
        try:
            res = await self.{model.__name__.lower()}_client.List(
                {self.app_name}_pb2.{model.__name__}ListRequest(), metadata=metadata
            )
            results = res.results
        except Exception as e:
            logger.error(f"Error retrieving {model.__name__.lower()}_name: {{e}}")
        if len(results) == 1 :
            return results[0].{model.__name__.lower()}_id
        else:
            raise ValueError(f"Error retrieving {model.__name__.lower()}_id - no or too many results found.")

    #@download_file
    #@download_image
    async def get_by_id(self, {model.__name__.lower()}_id: str = None) -> {self.base_app_name}_dm.{model.__name__}:
        \"\"\" retrieve a {model.__name__.lower()} data model by {model.__name__.lower()}_id \"\"\"
        try:
            res = await self.{model.__name__.lower()}_client.Retrieve(
                {self.app_name}_pb2.{model.__name__}RetrieveRequest({model.__name__.lower()}_id={model.__name__.lower()}_id)
            )
            return {self.base_app_name}_dm.{model.__name__}(**MessageToDict(res, preserving_proto_field_name=True))
        except Exception as e:
            logger.error(f"Error retrieving {model.__name__.lower()}_id: {{e}}")

    #@download_file
    #@download_image
    async def get_by_name(self, {model.__name__.lower()}_name: str = None,
                                namespace_uri : str = None) -> {self.base_app_name}_dm.{model.__name__}:
        \"\"\" retrieve a {model.__name__.lower()} data model by {model.__name__.lower()}_name \"\"\"
        if namespace_uri is None and self._default_namespace_uri is not None:
            namespace_uri = self._default_namespace_uri
        if namespace_uri is None:
            raise ValueError("No namespace URI is given.")
        try:
            namespace_id = await self.namespace_if.retrieve_id(namespace_uri=namespace_uri)
            res = await self.{model.__name__.lower()}_client.Retrieve(
                {self.app_name}_pb2.{model.__name__}RetrieveRequest(name={model.__name__.lower()}_name, namespace=namespace_id)
            )
            return {self.base_app_name}_dm.{model.__name__}(**MessageToDict(res, preserving_proto_field_name=True))
        except Exception as e:
            logger.error(f"Error retrieving {model.__name__.lower()}_name: {{e}}")
            
        
    async def list_items(self, filter_dict: dict = None, ids_only: bool = False) -> list[str]:
        \"\"\" list all {model.__name__.lower()}s \"\"\"
        if filter_dict is None:
            res = await self.{model.__name__.lower()}_client.List({self.app_name}_pb2.{model.__name__}ListRequest())
        else:
            metadata = (("filters", (json.dumps(filter_dict))),)
            res = await self.{model.__name__.lower()}_client.List(
                {self.app_name}_pb2.{model.__name__}ListRequest(), metadata=metadata
            )
        if res is not None and ids_only:
            return [item.{model.__name__.lower()}_id for item in res.results]
        else:
            return res.results
            

    async def list_full(self, filter_dict: dict = None) -> list:
        \"\"\" list all {model.__name__.lower()}s \"\"\"
        if self.{model.__name__.lower()}_full_client is None:
            raise ValueError("No full client available.")
        if filter_dict is None:
            return await self.{model.__name__.lower()}_full_client.List(
                {self.app_name}_pb2.{model.__name__}FullListRequest()
            )
        else:
            metadata = (("filters", (json.dumps(filter_dict))),)
            return await self.{model.__name__.lower()}_full_client.List(
                {self.app_name}_pb2.{model.__name__}FullListRequest(), metadata=metadata
            )

    @import_model_instance_from_file(relation_resolver=lambda x: x)
    # @update_file
    # @update_image
    async def update(self, {model.__name__.lower()} : {self.base_app_name}_dm.{model.__name__} = None) -> None:
        \"\"\" update a {model.__name__.lower()} model instance \"\"\"
        try:
            await self.{model.__name__.lower()}_client.Update(
                {self.app_name}_pb2.{model.__name__}Request(**{model.__name__.lower()}.model_dump()
            ))
        except Exception as e:
            logger.error(f"Error updating {model.__name__.lower()}_id: {{e}}")

    async def delete(self, id_list: list[str] = None, filter_dict: dict = None) -> None:
        \"\"\" delete a {model.__name__.lower()} by {model.__name__.lower()}_id or filter_dict \"\"\"
        if filter_dict is not None:
            id_list = await self.list_items(filter_dict=filter_dict, ids_only=True)
    
        try:
            delete_coroutines = ( self.{model.__name__.lower()}_client.Destroy({self.app_name}_pb2.{model.__name__}DestroyRequest({model.__name__.lower()}_id=id)) for id in id_list )
            await asyncio.gather(*delete_coroutines)
            
        except Exception as e:
            logger.error(f"Error deleting {model.__name__.lower()}_id: {{e}}")
\n"""
  


            


"""
________________________________________________________________________

:PROJECT: LARA

* LARA colour schemes*

:details:  colour schemes - 
            see: https://github.com/vaab/colour
            .

:author:  mark doerr <mark.doerr@uni.greifswald.de> 
          
:date: (creation)          20200522

.. note:: some remarks
.. todo:: - 
________________________________________________________________________

"""

__version__ = "v0.0.3"


from colour import Color

# def changeColorSaturation(color: str="greed", factor: int = 0):
#     return Color(color).lighter(factor).name(Color.HexRgb)


class ColorBlind4:
    """ source: https://davidmathlogic.com/colorblind/ 
    """
    magenta = "#D81B60"
    light_blue = "#1E88E5"
    light_orange = "#FFC107"
    dark_green = "#004D40"

    @property
    def full_scheme(self):
        return [self.magenta, self.light_blue, self.light_orange, self.dark_green]

    @property
    def magenta(self):
        return Color(self.magenta)

    @property
    def light_blue(self):
        return Color(self.light_blue)

    @property
    def light_orange(self):
        return Color(self.light_orange)

    @property
    def dark_green(self):
        return Color(self.dark_green)

    @property
    def full_color_scheme(self):
        return [Color(col) for col in self.full_scheme]

    # @property
    # def full_scheme_lighter(self, factor: int = 0):
    #     return [col.lighter(factor) for col in self.full_color_scheme]

    # @classmethod
    # def lighter_color(cls, color: Color = None, factor: int = 0):
    #     return color.lighter(factor)

    # @classmethod
    # def lighter_color_str(cls, color: Color = None, factor: int = 0):
    #     q_col = color.lighter(factor)
    #     return q_col.name(Color.HexRgb)


class ColorBlind8:
    """
        # 8- pfor deuteranopia  bs
        # 18:05:14 Tue 19 May 2020
        # Martin Krzywinski martink@bcgsc.ca
        # Methods and details: http://mkweb.bcgsc.ca/colorblind
        # Color names: http://mkweb.bcgsc.ca/colornames
        # Main palette colors are N-main with N-alt being (one of many) alternatives
        # that are indistinguishable to deuteranopes.
    """
    black = "#000000"  # 0   0   0 black, grey, grey, grey, rich black, grey, cod grey, grey, almost black, grey
    sapphire = "#2271B2"  # 34 113 178 honolulu blue, bluish, strong cornflower blue, spanish blue, medium persian blue, sapphire blue, ocean boat blue, french blue, windows blue, tufts blue
    purple = "#AA0DB4"  # 170  13 180 barney, strong magenta, heliotrope magenta, strong heliotrope, steel pink, barney purple, purple, violet, violet eggplant, deep magenta
    summer_sky = "#3DB7E9"  # 61 183 233 summer sky, cyan, picton blue, vivid cerulean, deep sky blue, brilliant cornflower blue, malibu, bright cerulean, cerulean, cerulean
    candy_pink = "#FF54ED"  # 255  84 237 light magenta, violet pink, light brilliant magenta, pink flamingo, light brilliant orchid, brilliant magenta, purple pizzazz, candy pink, blush pink, shocking pink
    wild_strawberry = "#F748A5"  # 247  72 165 barbie pink, rose bonbon, wild strawberry, brilliant rose, brilliant rose, magenta, wild strawberry, light brilliant rose, frostbite, brilliant cerise
    topaz = "#00B19F"  # 0 177 159 strong opal, tealish, persian green, keppel, topaz, manganese blue, light sea green, sea green light, puerto rico, turquoise
    ocean_green = "#359B73"  # 53 155 115 ocean green, sea green, viridian, mother earth, moderate spring green, moderate aquamarine, paolo veronese green, observatory, jungle green, ocean green
    vivid_rose = "#EB057A"  # 235   5 122 vivid rose, red purple, mexican pink, bright pink, rose, strong pink, luminous vivid rose, deep pink, winter sky, hot pink
    bamboo = "#d55e00"  # 213  94   0 bamboo, smoke tree, red stage, tawny, tenn, tenne, burnt orange, rusty orange, dark orange, mars yellow
    cherry_red = "#F8071D"  # 248   7  29 vivid red, luminous vivid amaranth, ruddy, ku crimson, vivid amaranth, light brilliant red, cherry red, red, red, bright red
    sun = "#e69f00"  # 230 159   0 gamboge, squash, buttercup, marigold, dark goldenrod, medium goldenrod, fuel yellow, sun, harvest gold, orange
    dark_orange = "#FF8D1A"  # 255 141  26 dark orange, juicy, west side, tangerine, gold drop, pizazz, princeton orange, university of tennessee orange, tangerine, tahiti gold
    # 240 228  66 holiday, buzz, paris daisy, starship, golden fizz, dandelion, gorse, lemon yellow, bright lights, sunflower
    lemon_yellow = "#f0e442"
    lime = "#9EFF37"  # 158 255  55 french lime, lime, green yellow, green lizard, luminous vivid spring bud, spring frost, vivid spring bud, bright yellow green, spring bud, acid green

    full_scheme = [black, sapphire, purple, summer_sky,
                   candy_pink, wild_strawberry, topaz,
                   ocean_green, vivid_rose, bamboo, cherry_red, sun, dark_orange, lemon_yellow, lime]

    # @property # wrong syntax !
    def full_color_scheme(self):
        return [Color(col) for col in self.full_scheme]

    # #@property wrong syntax
    # def full_scheme_lighter(self, factor: int = 30):
    #     return [col.lighter(factor) for col in self.full_color_scheme]

    # @classmethod
    # def lighter_qcolor(cls, color: Color = None, factor: int = 0):
    #     return color.lighter(factor)

    # @classmethod
    # def lighter_color_str(cls, color: Color = None, factor: int = 0):
    #     q_col = color.lighter(factor)
    #     return q_col.name(Color.HexRgb)


class ColorBlind16:
    """
        # 8- pfor deuteranopia  bs
        # 18:05:14 Tue 19 May 2020
        # Martin Krzywinski martink@bcgsc.ca
        # Methods and details: http://mkweb.bcgsc.ca/colorblind
        # Color names: http://mkweb.bcgsc.ca/colornames
        # Main palette colors are N-main with N-alt being (one of many) alternatives
        # that are indistinguishable to deuteranopes.
    """
    black = "#000000"  # 0   0   0 black, grey, grey, grey, rich black, grey, cod grey, grey, almost black, grey
    sapphire = "#2271B2"  # 34 113 178 honolulu blue, bluish, strong cornflower blue, spanish blue, medium persian blue, sapphire blue, ocean boat blue, french blue, windows blue, tufts blue
    purple = "#AA0DB4"  # 170  13 180 barney, strong magenta, heliotrope magenta, strong heliotrope, steel pink, barney purple, purple, violet, violet eggplant, deep magenta
    summer_sky = "#3DB7E9"  # 61 183 233 summer sky, cyan, picton blue, vivid cerulean, deep sky blue, brilliant cornflower blue, malibu, bright cerulean, cerulean, cerulean
    candy_pink = "#FF54ED"  # 255  84 237 light magenta, violet pink, light brilliant magenta, pink flamingo, light brilliant orchid, brilliant magenta, purple pizzazz, candy pink, blush pink, shocking pink
    wild_strawberry = "#F748A5"  # 247  72 165 barbie pink, rose bonbon, wild strawberry, brilliant rose, brilliant rose, magenta, wild strawberry, light brilliant rose, frostbite, brilliant cerise
    topaz = "#00B19F"  # 0 177 159 strong opal, tealish, persian green, keppel, topaz, manganese blue, light sea green, sea green light, puerto rico, turquoise
    ocean_green = "#359B73"  # 53 155 115 ocean green, sea green, viridian, mother earth, moderate spring green, moderate aquamarine, paolo veronese green, observatory, jungle green, ocean green
    vivid_rose = "#EB057A"  # 235   5 122 vivid rose, red purple, mexican pink, bright pink, rose, strong pink, luminous vivid rose, deep pink, winter sky, hot pink
    bamboo = "#d55e00"  # 213  94   0 bamboo, smoke tree, red stage, tawny, tenn, tenne, burnt orange, rusty orange, dark orange, mars yellow
    cherry_red = "#F8071D"  # 248   7  29 vivid red, luminous vivid amaranth, ruddy, ku crimson, vivid amaranth, light brilliant red, cherry red, red, red, bright red
    sun = "#e69f00"  # 230 159   0 gamboge, squash, buttercup, marigold, dark goldenrod, medium goldenrod, fuel yellow, sun, harvest gold, orange
    dark_orange = "#FF8D1A"  # 255 141  26 dark orange, juicy, west side, tangerine, gold drop, pizazz, princeton orange, university of tennessee orange, tangerine, tahiti gold
    # 240 228  66 holiday, buzz, paris daisy, starship, golden fizz, dandelion, gorse, lemon yellow, bright lights, sunflower
    lemon_yellow = "#f0e442"
    lime = "#9EFF37"  # 158 255  55 french lime, lime, green yellow, green lizard, luminous vivid spring bud, spring frost, vivid spring bud, bright yellow green, spring bud, acid green

    scheme8_str = [black, sapphire, purple, summer_sky,
                   candy_pink, wild_strawberry, topaz,
                   ocean_green, vivid_rose, bamboo, cherry_red, sun, dark_orange, lemon_yellow, lime]

    scheme8_color = [Color(col) for col in scheme8_str]

    # col.set_saturation(0.8)
    scheme8_light_color = [col for col in scheme8_color]

    full_color_scheme = scheme8_color + scheme8_light_color

    # @classmethod
    # def lighter_color(cls, color: Color = None, factor: float = 0.0):

    #     return color.set_saturation(factor)

    # @property # wrong syntax !
    # def full_color_scheme(self):
    #     return  full_color_scheme



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_base admin *

:details: lara_django_base admin module admin backend configuration.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev forms_generator -c lara_django_base > forms.py" to update this file
________________________________________________________________________
"""

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit
from django import forms

# from crispy_forms.bootstrap import FormActions
from .models import (
    Address,
    City,
    Country,
    CountrySubdivision,
    Currency,
    DataType,
    ExtraData,
    GeoLocation,
    ItemStatus,
    Location,
    MediaType,
    Namespace,
    PhysicalUnit,
    Room,
    RoomCategory,
    ScientificConstant,
    SubdivisionType,
    Tag,
)


class DataTypeCreateForm(forms.ModelForm):
    class Meta:
        model = DataType
        fields = ("name", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "iri", "description", Submit("submit", "Create"))


class DataTypeUpdateForm(forms.ModelForm):
    class Meta:
        model = DataType
        fields = ("name", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "iri", "description", Submit("submit", "Create"))


class MediaTypeCreateForm(forms.ModelForm):
    class Meta:
        model = MediaType
        fields = ("name", "media_type", "def_filename_extension_regex", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "media_type", "def_filename_extension_regex", "description", Submit("submit", "Create")
        )


class MediaTypeUpdateForm(forms.ModelForm):
    class Meta:
        model = MediaType
        fields = ("name", "media_type", "def_filename_extension_regex", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "media_type", "def_filename_extension_regex", "description", Submit("submit", "Create")
        )


class NamespaceCreateForm(forms.ModelForm):
    class Meta:
        model = Namespace
        fields = ("uri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("description", Submit("submit", "Create"))


class NamespaceUpdateForm(forms.ModelForm):
    class Meta:
        model = Namespace
        fields = ("uri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("description", Submit("submit", "Create"))


class TagCreateForm(forms.ModelForm):
    class Meta:
        model = Tag
        fields = ("tag",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("tag", Submit("submit", "Create"))


class TagUpdateForm(forms.ModelForm):
    class Meta:
        model = Tag
        fields = ("tag",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("tag", Submit("submit", "Create"))


class ItemStatusCreateForm(forms.ModelForm):
    class Meta:
        model = ItemStatus
        fields = ("status", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("status", "description", Submit("submit", "Create"))


class ItemStatusUpdateForm(forms.ModelForm):
    class Meta:
        model = ItemStatus
        fields = ("status", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("status", "description", Submit("submit", "Create"))


class ExtraDataCreateForm(forms.ModelForm):
    class Meta:
        model = ExtraData
        fields = ("name", "namespace", "data_json", "media_type", "iri", "url", "description", "file", "image")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "namespace",
            "data_json",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
            "image",
            Submit("submit", "Create"),
        )


class ExtraDataUpdateForm(forms.ModelForm):
    class Meta:
        model = ExtraData
        fields = ("name", "namespace", "data_json", "media_type", "iri", "url", "description", "file", "image")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "namespace",
            "data_json",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
            "image",
            Submit("submit", "Create"),
        )


class PhysicalUnitCreateForm(forms.ModelForm):
    class Meta:
        model = PhysicalUnit
        fields = ("name", "dimension", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "dimension", "description", Submit("submit", "Create"))


class PhysicalUnitUpdateForm(forms.ModelForm):
    class Meta:
        model = PhysicalUnit
        fields = ("name", "dimension", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "dimension", "description", Submit("submit", "Create"))


class ScientificConstantCreateForm(forms.ModelForm):
    class Meta:
        model = ScientificConstant
        fields = ("name", "value", "unit", "data_json", "iri", "url_source", "url_definition", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "value",
            "unit",
            "data_json",
            "iri",
            "url_source",
            "url_definition",
            "description",
            Submit("submit", "Create"),
        )


class ScientificConstantUpdateForm(forms.ModelForm):
    class Meta:
        model = ScientificConstant
        fields = ("name", "value", "unit", "data_json", "iri", "url_source", "url_definition", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "value",
            "unit",
            "data_json",
            "iri",
            "url_source",
            "url_definition",
            "description",
            Submit("submit", "Create"),
        )


class CurrencyCreateForm(forms.ModelForm):
    class Meta:
        model = Currency
        fields = ("name", "iri", "alpha3code", "numeric_code", "symbol", "exchange_rate_eur")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "iri", "alpha3code", "numeric_code", "symbol", "exchange_rate_eur", Submit("submit", "Create")
        )


class CurrencyUpdateForm(forms.ModelForm):
    class Meta:
        model = Currency
        fields = ("name", "iri", "alpha3code", "numeric_code", "symbol", "exchange_rate_eur")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "iri", "alpha3code", "numeric_code", "symbol", "exchange_rate_eur", Submit("submit", "Create")
        )


class GeoLocationCreateForm(forms.ModelForm):
    class Meta:
        model = GeoLocation
        fields = ("name", "coordinates_dd_lat", "coordinates_dd_long", "elevation", "openstreetmap", "googlemap")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "coordinates_dd_lat",
            "coordinates_dd_long",
            "elevation",
            "openstreetmap",
            "googlemap",
            Submit("submit", "Create"),
        )


class GeoLocationUpdateForm(forms.ModelForm):
    class Meta:
        model = GeoLocation
        fields = ("name", "coordinates_dd_lat", "coordinates_dd_long", "elevation", "openstreetmap", "googlemap")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "coordinates_dd_lat",
            "coordinates_dd_long",
            "elevation",
            "openstreetmap",
            "googlemap",
            Submit("submit", "Create"),
        )


class CountryCreateForm(forms.ModelForm):
    class Meta:
        model = Country
        fields = ("name", "name_local", "alpha2code", "alpha3code", "numeric_code", "phone_code", "icon", "currency")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "name_local",
            "alpha2code",
            "alpha3code",
            "numeric_code",
            "phone_code",
            "icon",
            "currency",
            Submit("submit", "Create"),
        )


class CountryUpdateForm(forms.ModelForm):
    class Meta:
        model = Country
        fields = ("name", "name_local", "alpha2code", "alpha3code", "numeric_code", "phone_code", "icon", "currency")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "name_local",
            "alpha2code",
            "alpha3code",
            "numeric_code",
            "phone_code",
            "icon",
            "currency",
            Submit("submit", "Create"),
        )


class SubdivisionTypeCreateForm(forms.ModelForm):
    class Meta:
        model = SubdivisionType
        fields = ("name", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "description", Submit("submit", "Create"))


class SubdivisionTypeUpdateForm(forms.ModelForm):
    class Meta:
        model = SubdivisionType
        fields = ("name", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "description", Submit("submit", "Create"))


class CountrySubdivisionCreateForm(forms.ModelForm):
    class Meta:
        model = CountrySubdivision
        fields = ("name", "name_local", "code", "alpha2code", "subdivision_type", "country", "timezone")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "name_local",
            "code",
            "alpha2code",
            "subdivision_type",
            "country",
            "timezone",
            Submit("submit", "Create"),
        )


class CountrySubdivisionUpdateForm(forms.ModelForm):
    class Meta:
        model = CountrySubdivision
        fields = ("name", "name_local", "code", "alpha2code", "subdivision_type", "country", "timezone")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "name_local",
            "code",
            "alpha2code",
            "subdivision_type",
            "country",
            "timezone",
            Submit("submit", "Create"),
        )


class CityCreateForm(forms.ModelForm):
    class Meta:
        model = City
        fields = ("name", "name_local", "country", "subdivision", "timezone")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_local", "country", "subdivision", "timezone", Submit("submit", "Create")
        )


class CityUpdateForm(forms.ModelForm):
    class Meta:
        model = City
        fields = ("name", "name_local", "country", "subdivision", "timezone")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_local", "country", "subdivision", "timezone", Submit("submit", "Create")
        )


class AddressCreateForm(forms.ModelForm):
    class Meta:
        model = Address
        fields = (
            
            "street",
            "house_number",
            "supplement",
            "city",
            "subdivision",
            "country",
            "postal_code",
            "geolocation",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            
            "street",
            "house_number",
            "supplement",
            "city",
            "subdivision",
            "country",
            "postal_code",
            "geolocation",
            Submit("submit", "Create"),
        )


class AddressUpdateForm(forms.ModelForm):
    class Meta:
        model = Address
        fields = (
            
            "street",
            "house_number",
            "supplement",
            "city",
            "subdivision",
            "country",
            "postal_code",
            "geolocation",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            
            "street",
            "house_number",
            "supplement",
            "city",
            "subdivision",
            "country",
            "postal_code",
            "geolocation",
            Submit("submit", "Create"),
        )


# class AddressCreateForm(forms.ModelForm):
#     class Meta:
#         model = Address
#         fields = ('building',
#                   'street',
#                   'house_number',
#                   'supplement',
#                   'city',
#                   'subdivision',
#                   'country',
#                   'postal_code',
#                   'geolocation',
#                   'data_extra')

#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.helper = FormHelper()
#         self.helper.layout = Layout('building',
#                                     Row(
#                                         Column(
#                                             'street', css_class='form-group col-md-6 mb-0'),
#                                         Column(
#                                             'house_number', css_class='form-group col-md-6 mb-0'),
#                                         Column(
#                                             'supplement', css_class='form-group col-md-6 mb-0'),
#                                         css_class='form-row'
#                                     ),
#                                     Row(
#                                         Column(
#                                             'postal_code',  css_class='form-group col-md-6 mb-0'),
#                                         Column(
#                                             'city', css_class='form-group col-md-6 mb-0'),
#                                         css_class='form-row'
#                                     ),
#                                     Row(
#                                         Column(
#                                             'subdivision',  css_class='form-group col-md-6 mb-0'),
#                                         Column(
#                                             'country', css_class='form-group col-md-6 mb-0'),
#                                         css_class='form-row'
#                                     ),
#                                     'geolocation',
#                                     'data_extra',
#                                     Submit(
#                                         'submit', 'Create'),
#                                     HTML(
#                                         '<a class="btn btn-danger" href="/people/addresses/list">Cancel</a>')
#                                     # Button('cancel', 'Cancel',
#                                     #        css_class='btn btn-danger')
#                                     )


# class AddressUpdateForm(forms.ModelForm):
#     class Meta:
#         model = Address
#         fields = ('building',
#                   'street',
#                   'house_number',
#                   'supplement',
#                   'city',
#                   'subdivision',
#                   'country',
#                   'postal_code',
#                   'geolocation',
#                   'data_extra')

#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.helper = FormHelper()
#         self.helper.layout = Layout('building',
#                                     Row(
#                                         Column(
#                                             'street', css_class='form-group col-md-6 mb-0'),
#                                         Column(
#                                             'house_number', css_class='form-group col-md-6 mb-0'),
#                                         Column(
#                                             'supplement', css_class='form-group col-md-6 mb-0'),
#                                         css_class='form-row'
#                                     ),
#                                     Row(
#                                         Column(
#                                             'postal_code',  css_class='form-group col-md-6 mb-0'),
#                                         Column(
#                                             'city', css_class='form-group col-md-6 mb-0'),
#                                         css_class='form-row'
#                                     ),
#                                     Row(
#                                         Column(
#                                             'subdivision',  css_class='form-group col-md-6 mb-0'),
#                                         Column(
#                                             'country', css_class='form-group col-md-6 mb-0'),
#                                         css_class='form-row'
#                                     ),
#                                     'geolocation',
#                                     'data_extra',
#                                     Submit(
#                                         'submit', 'Create')
#                                     )


class RoomCategoryCreateForm(forms.ModelForm):
    class Meta:
        model = RoomCategory
        fields = ("category", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("category", "description", Submit("submit", "Create"))


class RoomCategoryUpdateForm(forms.ModelForm):
    class Meta:
        model = RoomCategory
        fields = ("category", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("category", "description", Submit("submit", "Create"))


class RoomCreateForm(forms.ModelForm):
    class Meta:
        model = Room
        fields = (
            "name",
            "category",
            "number",
            "floor",
            "phone_number",
            "room_map",
            "max_people",
            "area",
            "volume",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "category",
            "number",
            "floor",
            "phone_number",
            "room_map",
            "max_people",
            "area",
            "volume",
            Submit("submit", "Create"),
        )


class RoomUpdateForm(forms.ModelForm):
    class Meta:
        model = Room
        fields = (
            "name",
            "category",
            "number",
            "floor",
            "phone_number",
            "room_map",
            "max_people",
            "area",
            "volume",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "category",
            "number",

            "phone_number",
            "room_map",
            "max_people",
            "area",
            "volume",
            Submit("submit", "Create"),
        )


class LocationCreateForm(forms.ModelForm):
    class Meta:
        model = Location
        fields = (
            "name",
            "barcode",
            "geolocation",
            "address",
            
            
           
            "location_json",
            "iri",
            "url",
            "description",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "barcode",
            "geolocation",
            "address",
            
            
           
            "location_json",
            "iri",
            "url",
            "description",
            Submit("submit", "Create"),
        )


class LocationUpdateForm(forms.ModelForm):
    class Meta:
        model = Location
        fields = (
            "name",
            "barcode",
            "geolocation",
            "address",
            
            
           
            "location_json",
            "iri",
            "url",
            "description",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "barcode",
            "geolocation",
            "address",
            
            
           
            "location_json",
            "iri",
            "url",
            "description",
            Submit("submit", "Create"),
        )


# from .forms import DataTypeCreateForm, MediaTypeCreateForm, NamespaceCreateForm, TagCreateForm, ItemStatusCreateForm,
#  ExtraDataCreateForm, PhysicalUnitCreateForm, ScientificConstantCreateForm, CurrencyCreateForm,
#  GeoLocationCreateForm, CountryCreateForm, SubdivisionTypeCreateForm, CountrySubdivisionCreateForm, CityCreateForm,
#  AddressCreateForm, RoomCategoryCreateForm, RoomCreateForm, LocationCreateFormDataTypeUpdateForm, MediaTypeUpdateForm,
#  NamespaceUpdateForm, TagUpdateForm, ItemStatusUpdateForm, ExtraDataUpdateForm, PhysicalUnitUpdateForm,
# ScientificConstantUpdateForm, CurrencyUpdateForm, GeoLocationUpdateForm, CountryUpdateForm,
# SubdivisionTypeUpdateForm, CountrySubdivisionUpdateForm, CityUpdateForm, AddressUpdateForm,
# RoomCategoryUpdateForm, RoomUpdateForm, LocationUpdateForm



# -*- coding: utf-8 -*-
"""_____________________________________________________________________

:PROJECT: LARAsuite

*generate unittests *

:details: generate_grpc_client - a unittests generator for LARA-django.
          This is a LARA-django-base management command to generate a unittests for a
          given app. The generated client is a python module that can be used to
          communicate with the gRPC server of the given app.

          requirements:
                python-lorem
:usage: lara-django-dev generate_unittests --author-name "mark doerr" --author-email "mark.doerr@uni-greifswald.de" --force lara_django_base

:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - https://stackoverflow.com/questions/11380413/python-unittest-passing-arguments
          - speed improvemnts https://medium.com/@thehackadda/tips-for-speeding-up-your-django-tests-35bab8250760
________________________________________________________________________
"""

import re
import os
import logging
import lorem  # pip install python-lorem
import random
import uuid
from datetime import datetime
import json

from django.apps import apps
from django.conf import settings
from django.core.management.base import LabelCommand, CommandError
from django.db import models


# Configurable constants
MAX_LINE_WIDTH = getattr(settings, "MAX_LINE_WIDTH", 120)
INDENT_WIDTH = getattr(settings, "INDENT_WIDTH", 4)

UNITTEST_FILE_HEADER = """\"\"\"_____________________________________________________________________

:PROJECT: LARAsuite

*{lara_app} unittests *

:details: {lara_app} unittests.
         - generated with 'lara-django-dev generate_unittests {lara_app}'
:authors: {lara_author} <{lara_author_email}>

.. note:: -
.. todo:: - 
________________________________________________________________________
\"\"\""""


def indent(level=1, text="", end="\n"):
    return "\n".join([(level * INDENT_WIDTH * " ") + line for line in text.split("\n")]) + end


def lower_strip(s):
    return s.lower().replace(" ", "").strip()


class Command(LabelCommand):
    help = """Generate a unittests for a given LARA app (models) - 1a"""
    args = "[app_name, author_name, author_email]"
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument("app_name", help="Name of the LARA app to generate the unittests for")

        parser.add_argument("--author-name", help="Name of the author", default="lara")
        parser.add_argument("--author-email", help="Email of the author", default="lara@lara.org")
        parser.add_argument("--force", action="store_true", help="overwrite existing files", default=False)
        # parser.add_argument('model_name', nargs='*')

    # @signalcommand
    def handle(self, *args, **options):
        self.app_name = options["app_name"]

        logging.warning("!! only a scaffold is generated, please check/add content to the generated files !!!")

        try:
            app = apps.get_app_config(self.app_name)
        except LookupError:
            self.stderr.write("This command requires an existing app name as argument")
            self.stderr.write("Available apps:")
            app_labels = [app.label for app in apps.get_app_configs()]
            for label in sorted(app_labels):
                self.stderr.write("    %s" % label)
            return

        model_res = []
        # for arg in options['model_name']:
        #     model_res.append(re.compile(arg, re.IGNORECASE))

        AppUnittestGenerator(app, model_res, options)

        # self.stdout.write()


class AppUnittestGenerator:
    def __init__(self, app_config, model_res, options):
        self.app_config = app_config
        self.model_res = model_res

        self.app_name = options["app_name"]
        self.author_name = options["author_name"]
        self.author_email = options["author_email"]
        self.force_overwrite = options["force"]

        self.model_dict = {}
        self.internal_model_name = set()
        self.external_model_name = set()

        self.unittests_str = ""
        self.model_names = [model.__name__ for model in self.app_config.get_models()]

        self.generate_unittests()

    def generate_unittests(self):
        self.generate_unittests_header()
        self.generate_unittests_imports()
        self.generate_model_tests()
        self.generate_unittests_main()

        # save unittests to file
        unittest_filename = f"test_{self.app_name}.py"

        print(f"writing {unittest_filename} to {os.getcwd()}")
        print(self.unittests_str)

        if os.path.isfile(unittest_filename):
            if self.force_overwrite:
                with open(unittest_filename, "w") as f:
                    f.write(self.unittests_str)
            else:
                append = input(
                    f"{unittest_filename} already exists,\n press 'a' to append,\n press 'o' to overwrite,\n or 'n' to do nothing: "
                )
                if append.lower() == "a":
                    with open(unittest_filename, "a") as f:
                        f.write(self.unittests_str)
                elif append.lower() == "o":
                    with open(unittest_filename, "w") as f:
                        f.write(self.unittests_str)
        else:
            with open(unittest_filename, "w") as f:
                f.write(self.unittests_str)

    def generate_unittests_header(self):
        self.unittests_str += UNITTEST_FILE_HEADER.format(
            lara_app=self.app_name,
            lara_author=self.author_name,
            lara_author_email=self.author_email,
        )

    def generate_unittests_imports(self):

        self.unittests_str += f"""

#import asyncio
import json
import logging
import uuid
from unittest import skip
from django.test import TestCase
from django.utils import timezone
from django.conf import settings


from {self.app_name}.models import ( {', '.join(self.model_names)}\n)\n

TEST_NUM_INSTANCES = 1  # TODO: set the number via command line argument\n\n
"""

    def generate_model_tests(self):

        unittests_handler_str = ""

        for model in self.app_config.get_models():

            self.unittests_str += f"""#_________________  {model.__name__}  ______________________\n

@skip("skip {model.__name__} tests, because they are not implemented yet.")
class {model.__name__.capitalize()}CRUDTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        # setting up the model data in the test database
        cls.{model.__name__.lower()}_model = [
            {model.__name__}.create({ self.parse_fields_for_init(model) }
        ) for i in range(TEST_NUM_INSTANCES) ]


    def setUp(self):
        # setting up the test environment
        pass

     def test_create_{model.__name__.lower()}(self):
        # testing creating objects from the {model.__name__} model

        #cat = Animal.objects.get(name="cat")
        #self.assertEqual(lion.speak(), 'The lion says "roar"')
        pass
              
    def test_retrieve_{model.__name__.lower()}(self):
        # testing retrieving objects from the {model.__name__} model

        #cat = Animal.objects.get(name="cat")
        #self.assertEqual(lion.speak(), 'The lion says "roar"')
        pass
    
    def test_update_{model.__name__.lower()}(self):
        # testing updating objects from the {model.__name__} model

        #self.assertEqual(lion.speak(), 'The lion says "roar"')
        pass

    def test_delete_{model.__name__.lower()}(self):
        # testing deleting objects from the {model.__name__} model
        pass\n\n"""

    def parse_fields_for_init(self, model):
        fields_str = "\n"

        default_prefix = "lara"

        for field in model._meta.get_fields():
            # print(f"field: {field} --> {type(field)} \n")

            if (
                isinstance(field, models.fields.related.ManyToManyField)
                or isinstance(field, models.fields.reverse_related.ManyToOneRel)
                or isinstance(field, models.fields.reverse_related.ManyToManyRel)
                or isinstance(field, models.fields.reverse_related.OneToOneRel)
                or isinstance(field, models.fields.reverse_related.ManyToManyRel)
            ):
                required = False
            else:
                required = field.blank == False

            match field:
                case models.fields.related.ManyToOneRel():
                    pass
                case models.fields.UUIDField():
                    if field.primary_key:
                        fields_str += indent(4, f"# {field.name} =    # (primary key) {field.help_text}")
                    else:
                        fields_str += indent(4, f"{field.name} = {uuid.uuid4()},   # (UUID) {field.help_text}")
                    # self.external_model_name.add("UUID")

                case models.fields.BooleanField():
                    fields_str += indent(
                        4, f"{field.name} = {random.choice(('True', 'False') ) },   # (bool) {field.help_text}"
                    )
                case models.fields.URLField():
                    url_str = f"http://www.{lorem.get_word()}.org"
                    if field.name == "url":
                        url_str = f"http://www.{lorem.get_word()}.org"
                    elif field.name == "iri":
                        url_str = f"http://w3id.org/lara/{model.__name__}_{{i}}"

                    fields_str += indent(4, f'{field.name} = f"{url_str}",   # (URL - str) {field.help_text}')
                case models.fields.CharField():
                    fields_str += indent(4, f'{field.name} = "{lorem.get_word()}",   # (str) {field.help_text}')

                case models.fields.TextField():
                    if field.name == "name" or field.name == "title":
                        fields_str += indent(4, f'{field.name} = f"{lorem.get_word()}_{{i}}",   # (str) {field.help_text}')
                    else:
                        fields_str += indent(4, f'{field.name} = "{lorem.get_sentence()}",   # (str) {field.help_text}')

                case models.fields.IntegerField():
                    fields_str += indent(4, f"{field.name} = {random.randint(1,100)},   # (int) {field.help_text}")

                case models.fields.FloatField():

                    fields_str += indent(
                        4, f"{field.name} = {random.uniform(1.0, 100.0)},   # (float) {field.help_text}"
                    )

                case models.fields.DecimalField():
                    fields_str += indent(
                        4, f"{field.name} = {random.uniform(1.0, 100.0)},   # (float) {field.help_text}"
                    )

                case models.fields.DateTimeField():
                    fields_str += indent(4, f'{field.name} = "{datetime.now()}",   # (datetime) {field.help_text}')

                case models.fields.DateField():
                    fields_str += indent(4, f'{field.name} = "{datetime.now().date()}",   # (date) {field.help_text}')

                case models.fields.json.JSONField():
                    # random json
                    rand_json = json.dumps(
                        {
                            "key1": lorem.get_word(),
                            "key2": lorem.get_word(),
                            "key3": lorem.get_word(),
                        }
                    )
                    fields_str += indent(4, f"{field.name} = {rand_json},   # (str) {field.help_text}")

                case models.fields.files.FileField():
                    fields_str += indent(4, f'{field.name} = "my_file.md",  # (file) {field.help_text}')

                case models.fields.related.ForeignKey():
                    # print(f"{field.name} --- attributes: {field.__dict__}")
                    fields_str += indent(
                        4, f"{field.name} =  ,  # (ForeignKey ->  {field.related_model.__name__}) {field.help_text}"
                    )
                    self.external_model_name.add(field.related_model.__name__)

                    self.model_dict[field.related_model.__name__] = field.related_model

                case models.OneToOneField():
                    fields_str += indent(
                        4, f"{field.name} =  ,  # (OneToOneField -> {field.related_model.__name__}) {field.help_text}"
                    )
                    self.external_model_name.add(field.related_model.__name__)

                    self.model_dict[field.related_model.__name__] = field.related_model

                case models.fields.related.ManyToManyField():
                    # print(f"field attributes: {field.__dict__}")
                    fields_str += indent(
                        4, f"{field.name} =  ,  # (ManyToManyField -> {field.related_model.__name__}) {field.help_text}"
                    )
                    self.external_model_name.add(field.related_model.__name__)

                    self.model_dict[field.related_model.__name__] = field.related_model

        return fields_str

    def get_relational_fields(self, model):
        foreignkey_fields = set()
        related_names = set()
        m2m_fields = set()
        one2one_fields = set()

        for field in model._meta.get_fields():
            print(f"field: {field} --> {type(field)} ")
            if isinstance(field, models.ForeignKey):
                foreignkey_fields.add(field.name)
            if isinstance(field, models.fields.related.ForeignKey):
                pass
                # print(f"field attributes: {field.__dict__}")
                # related_names.add(field.name)
            if isinstance(field, models.fields.related.ManyToManyField):
                # print(f"field attributes: {field.__dict__}")
                m2m_fields.add(field.name)
            if isinstance(field, models.fields.reverse_related.ManyToManyRel):
                # print(f"field attributes: {field.__dict__}")
                related_names.add(field.name)
                related_names.add(field.related_name)
                related_names.add(field.related_query_name)
                related_names.add(field.field.name)
                # related_names.add(field.field.related_name)
                related_names.add(field.field.related_query_name)
            if isinstance(field, models.fields.related.OneToOneField):
                one2one_fields.add(field.name)
            if isinstance(field, models.fields.reverse_related.OneToOneRel):
                related_names.add(field.name)
                related_names.add(field.related_name)
                related_names.add(field.related_query_name)
        return [foreignkey_fields, m2m_fields, one2one_fields, related_names]

    def generate_unittests_main(self):
        self.unittests_str += f"""
# async def main():
    # (async) main function
    

if __name__ == "__main__":
    # parse command line arguments
    import argparse

    # add num test instances as command line argument
    parser = argparse.ArgumentParser(description="unittests for {self.app_name}")
    parser.add_argument("--verbose", help="verbose output", action="store_true")
    parser.add_argument("--num-instances", help="number of test instances", type=int, default=1)

    args = parser.parse_args()

    # set the number of test instances
    TEST_NUM_INSTANCES = args.num_instances

    # run the unittests
    # asyncio.run(main())
"""



from django.conf import settings


def external_services_urls(request):
    return {
        "LARA_DOCS_URL": settings.LARA_DOCS_URL,
        "PREFECT_DASHBOARD_URL" : settings.PREFECT_DASHBOARD_URL,
        "VIRTUOSO_SPARQL_URL": settings.VIRTUOSO_SPARQL_URL,
        "MINIO_DASHBOARD" : settings.MINIO_DASHBOARD,
    }